"""Module for running code on application exit"""

import atexit
import signal
from typing import Callable

from . import error, log
from ._storage import InternalStorage

logger = log.get_logger("onexit")

def register(name: str, callback: Callable) -> None:
    """
    Run the given callback regardless of how the application exits.
    
    Does not work if the application is killed with SIGKILL.
    """

    if "." in name:
        name = name.replace(".", "_")

    registered = False

    if f"_onexit.atexit.{name}" not in InternalStorage:
        register_normal_exit(name, callback)
        registered = True

    if f"_onexit.signal.{name}" not in InternalStorage:
        register_sigterm(name, callback)
        registered = True

    if not registered:
        raise error.RegisteredMultipleTimesError.with_values(name)

def register_normal_exit(name: str, callback: Callable) -> None:
    """
    Run the given callback if the application exits normally or with an exception.
    
    Does not work if the application is killed with SIGTERM or SIGKILL.
    """

    if "." in name:
        name = name.replace(".", "_")

    if f"_onexit.atexit.{name}" in InternalStorage:
        raise error.RegisteredMultipleTimesError.with_values(name)

    InternalStorage[f"_onexit.atexit.{name}"] = callback

def register_sigterm(name: str, callback: Callable) -> None:
    """
    Run the given callback if the application is killed with SIGTERM.
    
    Does not work if the application exits normally, exits with an exception or is killed with SIGKILL.
    """

    _ensure_signal_handler()

    if "." in name:
        name = name.replace(".", "_")

    if f"_onexit.signal.{name}" in InternalStorage:
        raise error.RegisteredMultipleTimesError.with_values(name)

    InternalStorage[f"_onexit.signal.{name}"] = callback

    # set handler if first signal callback is registered
    if len(InternalStorage["_onexit.signal"]) == 1:
        signal.signal(signal.SIGTERM, _signal_func)

def deregister(name: str) -> None:
    """
    Deregister the callback with the given name.
    
    Raises an NameNotFoundError if the name is not yet registered.
    """

    deleted = False

    if f"_onexit.atexit.{name}" in InternalStorage:
        deregister_normal_exit(name)
        deleted = True
    if f"_onexit.signal.{name}" in InternalStorage:
        deregister_sigterm(name)
        deleted = True

    if not deleted:
        # no callback was deleted
        raise error.NameNotFoundError.with_values(name)

def deregister_normal_exit(name: str) -> None:
    """
    Deregister the callback with the given name.
    
    Raises an NameNotFoundError if the name is not yet registered.
    """

    if f"_onexit.atexit.{name}" not in InternalStorage:
        raise error.NameNotFoundError.with_values(name)

    del InternalStorage[f"_onexit.atexit.{name}"]

def deregister_sigterm(name: str) -> None:
    """
    Deregister the callback with the given name.
    
    Raises an NameNotFoundError if the name is not yet registered.
    """

    _ensure_signal_handler()

    if f"_onexit.signal.{name}" not in InternalStorage:
        raise error.NameNotFoundError.with_values(name)

    del InternalStorage[f"_onexit.signal.{name}"]

    # reset handler if no signal callback is registered anymore
    if "_onexit.signal" not in InternalStorage:
        signal.signal(signal.SIGTERM, InternalStorage["_onexit.orig.signal"])

def reset() -> None:
    """Reset all regsitered callbacks"""

    if "_onexit.atexit" in InternalStorage:
        for name in list(InternalStorage["_onexit.atexit"].keys()):
            deregister_normal_exit(name)

    if "_onexit.signal" in InternalStorage:
        for name in list(InternalStorage["_onexit.signal"].keys()):
            deregister_sigterm(name)

def _atexit_func():
    if "_onexit.atexit" not in InternalStorage:
        return

    for callback in InternalStorage["_onexit.atexit"].values():
        try:
            callback()
        # pylint: disable-next=broad-exception-caught
        except Exception as e:
            logger.exception(e)

# pylint: disable-next=unused-argument
def _signal_func(signum, frame):
    if "_onexit.signal" not in InternalStorage:
        return

    for callback in InternalStorage["_onexit.signal"].values():
        try:
            callback()
        # pylint: disable-next=broad-exception-caught
        except Exception as e:
            logger.exception(e)

def _ensure_signal_handler():
    if "_onexit.signal" not in InternalStorage:
        if signal.getsignal(signal.SIGTERM) is not InternalStorage["_onexit.orig.signal"]:
            raise RuntimeError("signal handler was overwritten, "
                               + "make sure to only use this module to set signal handlers")
    else:
        if signal.getsignal(signal.SIGTERM) is not _signal_func:
            raise RuntimeError("signal handler was overwritten, "
                               + "make sure to only use this module to set signal handlers")

InternalStorage["_onexit.orig.signal"] = signal.getsignal(signal.SIGTERM)

atexit.register(_atexit_func)



"""Module containing the _BaseStorage base class"""

from __future__ import annotations

from typing import Any

from .. import error

class _AutoremoveDict(dict):
    """An internal class representing auto-removable subdicts within the storage"""

class _BaseStorage():
    def __init__(self) -> None:
        raise NotImplementedError()

    _instance: _BaseStorage = None
    _store: dict[str, Any] = None

    _STORAGE_NAME = "BaseStorage"

    @property
    def name(self) -> str:
        """Provide a human-readable name for this storage"""

        return self._STORAGE_NAME

    @name.setter
    def name(self, _) -> None:
        raise error.ReadonlyError.with_values("Storage.name")

    def contains_item(self, key: str, item: Any) -> bool:
        """
        Check whether a key within the storage equals an item.

        If 'key' contains a '.', also checks if all sub-dicts exist.
        """

        if not self._contains(key):
            return False
        return item == self[key]

    def contains(self, key: str) -> bool:
        """
        Check whether a key exists within the storage.

        If 'key' contains a '.', also checks if all sub-dicts exist.
        """

        return self._contains(key)

    def get(self, key: str, default: Any = None) -> Any:
        """
        Return the value of an key if it exists in the storage.

        If the key is not found, return the default value instead.
        """

        if self._contains(key):
            return self._get(key)

        return default

    def items(self):
        """
        Return a view on the top-level keys and values in the storage.
        """

        return self._store.items()

    def pop(self, key: str) -> Any:
        """
        Return the value of an key if it exists in the storage.
        """

        val = self._get(key)
        self._del(key)
        return val

    def keys(self):
        """
        Return a view on the top-level keys in the storage.
        """

        self._ensure_initialized()

        return self._store.keys()

    def values(self):
        """
        Return a view on the top-level items in the storage.
        """

        self._ensure_initialized()

        return self._store.values()

    def __getitem__(self, key: str) -> Any:
        return self._get(key)

    def __setitem__(self, key: str, item: Any) -> None:
        return self._set(key, item)

    def __delitem__(self, key: str) -> None:
        return self._del(key)

    def __contains__(self, key: str) -> bool:
        return self._contains(key)

    def __str__(self) -> str:
        return str(self._store)

    def _contains(self, key: str) -> bool:
        self._ensure_initialized()
        self._ensure_key_validity(key)

        if "." not in key:
            return key in self._store

        parts = key.split(".")
        curr_dict: dict[str, Any] = self._store
        for c, part in enumerate(parts):
            if part not in curr_dict:
                return False
            # if it isn't the last part
            if c < len(parts) - 1:
                curr_dict: dict[str, Any] = curr_dict[part]

        return parts[-1] in curr_dict

    def _get(self, key: str) -> Any:
        self._ensure_initialized()
        self._ensure_key_validity(key)

        if "." not in key:
            if key not in self._store:
                raise error.KeyNotFoundError.with_values(key)
            return self._store[key]

        parts = key.split(".")
        curr_dict = self._store
        for c, part in enumerate(parts):
            if part not in curr_dict:
                invalid_key = ""
                for item in parts:
                    invalid_key += f"{item}" if invalid_key == "" else f".{item}"
                    if item == part:
                        break
                raise error.KeyNotFoundError.with_values(invalid_key)
            # if it isn't the last part
            if c < len(parts) - 1:
                curr_dict = curr_dict[part]

        return curr_dict[parts[-1]]

    def _set(self, key: str, item: Any) -> None:
        self._ensure_initialized()
        self._ensure_key_validity(key)

        if "." not in key:
            self._store[key] = item
            return

        parts = key.split(".")
        curr_dict = self._store
        for c, part in enumerate(parts):
            # if it isn't the last part
            if c < len(parts) - 1:
                # add a missing dictionary
                if part not in curr_dict:
                    curr_dict[part] = _AutoremoveDict()
                curr_dict = curr_dict[part]
            else:
                # add the actual item
                curr_dict[part] = item

    def _del(self, key: str) -> None:
        self._ensure_initialized()
        self._ensure_key_validity(key)

        if "." not in key:
            if key not in self._store:
                raise error.KeyNotFoundError.with_values(key)
            del self._store[key]
            return

        parts = key.split(".")
        curr_dict = self._store
        for c, part in enumerate(parts):
            if part not in curr_dict:
                invalid_key = ""
                for item in parts:
                    invalid_key += f"{item}" if invalid_key == "" else f".{item}"
                    if item == part:
                        break
                raise error.KeyNotFoundError.with_values(invalid_key)

            # if it isn't the last part
            if c < len(parts) - 1:
                curr_dict = curr_dict[part]
            else:
                # delete the actual item
                del curr_dict[part]

        # delete empty autogenerated dicts
        key = key.rsplit(".", maxsplit=1)[0]
        parts = key.split(".")

        while True:
            curr_dict = self._store

            for c, part in enumerate(parts):
                # if it isn't the last part
                if c < len(parts) - 1:
                    curr_dict = curr_dict[part]
                else:
                    # pylint: disable-next=unidiomatic-typecheck
                    if type(curr_dict[part]) == _AutoremoveDict and len(curr_dict[part]) == 0:
                        # delete the actual item
                        del curr_dict[part]
                    else:
                        # we are done
                        return

            if "." in key:
                key = key.rsplit(".", maxsplit=1)[0]
                parts = key.split(".")
            else:
                # we deleted every subkey
                return

    def _ensure_initialized(self) -> None:
        if self._store is None:
            raise error.NotInitializedError()

    def _ensure_key_validity(self, key: Any) -> None:
        if not isinstance(key, str):
            raise error.WrongTypeError.with_values(key, str)

        if "." in key:
            if key[0] == ".":
                raise error.InvalidKeyError("Key cannot start with '.'")
            if key[-1] == ".":
                raise error.InvalidKeyError("Key cannot end with '.'")
            if ".." in key:
                raise error.InvalidKeyError("Key cannot contain '..'")

    def __init_subclass__(cls):
        if cls._STORAGE_NAME == "BaseStorage":
            raise error.UninitializedFieldError.with_values(cls, "_STORAGE_NAME")

        if not isinstance(cls._STORAGE_NAME, str):
            raise error.WrongTypeError.with_values(cls._STORAGE_NAME, str)



"""Module containing tests for the abllib.wrapper.deprecated functionality"""

# pylint: disable=function-redefined, consider-using-with, unused-argument, missing-class-docstring

import pytest

from abllib import error, wrapper

def test_callable():
    """Ensure that wrapper.deprecated is callable"""

    assert hasattr(wrapper, "deprecated")
    assert callable(wrapper.deprecated)

def test_defaultmessage(capture_logs):
    """Ensure that wrapper.deprecated logs a warning"""

    @wrapper.deprecated
    def func1():
        pass
    func1()
    with open("test.log", "r", encoding="utf8") as f:
        assert "The functionality 'func1' is deprecated" in f.read()

def test_defaultmessage_class(capture_logs):
    """Ensure that wrapper.deprecated logs a warning"""

    @wrapper.deprecated
    class A():
        pass
    A()
    with open("test.log", "r", encoding="utf8") as f:
        assert "The functionality 'A' is deprecated" in f.read()

def test_warning_defaultmessage(capture_logs):
    """Ensure that wrapper.deprecated.warning logs a warning"""

    @wrapper.deprecated.warning
    def func1():
        pass
    func1()
    with open("test.log", "r", encoding="utf8") as f:
        assert "The functionality 'func1' is deprecated" in f.read()

def test_warning_defaultmessage_class(capture_logs):
    """Ensure that wrapper.deprecated.warning logs a warning"""

    @wrapper.deprecated.warning
    class A():
        pass
    A()
    with open("test.log", "r", encoding="utf8") as f:
        assert "The functionality 'A' is deprecated" in f.read()

def test_custommessage(capture_logs):
    """Ensure that wrapper.deprecated with a custom message logs that message"""

    @wrapper.deprecated("A custom deprecation message")
    def func1():
        pass
    func1()
    with open("test.log", "r", encoding="utf8") as f:
        assert "A custom deprecation message" in f.read()

def test_custommessage_class(capture_logs):
    """Ensure that wrapper.deprecated with a custom message logs that message"""

    @wrapper.deprecated("A custom deprecation message")
    class A():
        pass
    A()
    with open("test.log", "r", encoding="utf8") as f:
        assert "A custom deprecation message" in f.read()

def test_warning_custommessage(capture_logs):
    """Ensure that wrapper.deprecated.warning with a custom message logs that message"""

    @wrapper.deprecated.warning("A custom deprecation message")
    def func1():
        pass
    func1()
    with open("test.log", "r", encoding="utf8") as f:
        assert "A custom deprecation message" in f.read()

def test_warning_custommessage_class(capture_logs):
    """Ensure that wrapper.deprecated.warning with a custom message logs that message"""

    @wrapper.deprecated.warning("A custom deprecation message")
    class A():
        pass
    A()
    with open("test.log", "r", encoding="utf8") as f:
        assert "A custom deprecation message" in f.read()

def test_reassign_func_defaultmessage(capture_logs):
    """Ensure that reassigning a function to a different name logs that name"""

    def func1():
        pass
    def func2():
        return func1()
    func2 = wrapper.deprecated.warning(func2)
    func2()
    with open("test.log", "r", encoding="utf8") as f:
        assert "The functionality 'func2' is deprecated" in f.read()

def test_inherit_class_defaultmessage(capture_logs):
    """Ensure that reassigning a function to a different name logs that name"""

    class A():
        pass
    @wrapper.deprecated.warning
    class B(A):
        pass
    B()
    with open("test.log", "r", encoding="utf8") as f:
        assert "The functionality 'B' is deprecated" in f.read()

def test_error_defaultmessage():
    """Ensure that wrapper.deprecated.error raises an error"""

    @wrapper.deprecated.error
    def func1():
        pass
    with pytest.raises(error.DeprecatedError):
        func1()

def test_error_custommessage():
    """Ensure that wrapper.deprecated as error with a custom message uses that message"""

    @wrapper.deprecated("A custom deprecation message", True)
    def func1():
        pass
    with pytest.raises(error.DeprecatedError):
        func1()

def test_error_custommessage():
    """Ensure that wrapper.deprecated.error with a custom message uses that message"""

    @wrapper.deprecated.error("A custom deprecation message")
    def func1():
        pass
    with pytest.raises(error.DeprecatedError):
        func1()



"""Module containing tests for abllib.thread"""

# pylint: disable=missing-class-docstring

from multiprocessing import Process
from threading import Thread
from time import sleep

import pytest

from abllib import pproc, wrapper

def test_workerthread_inheritance():
    """Ensure that WorkerThread inherits from Thread"""

    assert hasattr(pproc, "WorkerThread")
    assert issubclass(pproc.WorkerThread, Thread)

def test_workerthread_func_execution():
    """Ensure that WorkerThread executes its target and then exits"""

    out = [False]
    def func1():
        out[0] = True

    t = pproc.WorkerThread(target=func1)
    t.start()

    c = 0
    while c < 10 and not out[0] and t.is_alive():
        sleep(0.1)
        c += 1

    if c >= 10:
        pytest.fail("thread did not complete in time")

    # is_alive stays True for a short time after thread completes
    sleep(0.1)
    assert not t.is_alive()

    assert out[0]

def test_workerthread_value_return():
    """Ensure that WorkerThread returns values"""

    def func1():
        return None

    t = pproc.WorkerThread(target=func1)
    t.start()

    r = t.join()
    assert r is None
    r = t.join(reraise=True)
    assert r is None

    def func2():
        return ("val1", 25)

    t = pproc.WorkerThread(target=func2)
    t.start()

    r = t.join()
    assert not t.is_alive()
    assert not t.failed()
    assert r == ("val1", 25)
    assert isinstance(r[0], str)
    assert isinstance(r[1], int)

    r = t.join(reraise=True)
    assert not t.is_alive()
    assert not t.failed()
    assert r == ("val1", 25)
    assert isinstance(r[0], str)
    assert isinstance(r[1], int)


def test_workerthread_exception_return():
    """Ensure that WorkerThread returns exceptions"""

    def func1():
        raise AssertionError("This is a test message")

    t = pproc.WorkerThread(target=func1)
    t.start()

    r = t.join()
    assert t.failed()
    assert isinstance(r, BaseException)
    assert str(r) == "This is a test message"

def test_workerthread_exception_reraise():
    """Ensure that WorkerThread reraises exceptions if flag is set"""

    def func1():
        raise AssertionError("This is a test message")

    t = pproc.WorkerThread(target=func1)
    t.start()

    try:
        t.join(reraise=True)
    except AssertionError as e:
        assert str(e) == "This is a test message"
    else:
        pytest.fail("no exception raised")

def test_workerprocess_inheritance():
    """Ensure that WorkerProcess inherits from Process"""

    assert hasattr(pproc, "WorkerProcess")
    assert issubclass(pproc.WorkerProcess, Process)

def test_workerprocess_value_return():
    """Ensure that WorkerProcess returns values"""

    def func1():
        return None

    p = pproc.WorkerProcess(target=func1)
    p.start()

    r = p.join()
    assert r is None
    r = p.join(reraise=True)
    assert r is None

    def func2():
        return ("val1", 25)

    p = pproc.WorkerProcess(target=func2)
    p.start()

    r = p.join()
    assert not p.failed()
    assert r == ("val1", 25)
    assert isinstance(r[0], str)
    assert isinstance(r[1], int)

    r = p.join(reraise=True)
    assert not p.failed()
    assert r == ("val1", 25)
    assert isinstance(r[0], str)
    assert isinstance(r[1], int)

def test_workerprocess_exception_return():
    """Ensure that WorkerProcess returns exceptions"""

    def func1():
        raise AssertionError("This is a test message")

    p = pproc.WorkerProcess(target=func1)
    p.start()

    r = p.join()
    assert p.failed()
    assert isinstance(r, BaseException)
    assert str(r) == "This is a test message"

def test_workerprocess_exception_reraise():
    """Ensure that WorkerProcess reraises exceptions if flag is set"""

    def func1():
        raise AssertionError("This is a test message")

    t = pproc.WorkerProcess(target=func1)
    t.start()

    try:
        t.join(reraise=True)
    except AssertionError as e:
        assert str(e) == "This is a test message"
    else:
        pytest.fail("no exception raised")

def test_lock():
    """Ensure that Lock is imported correctly"""

    assert hasattr(wrapper, "Lock")
    assert pproc.Lock == wrapper.Lock

def test_semaphore():
    """Ensure that Semaphore is imported correctly"""

    assert hasattr(wrapper, "Semaphore")
    assert pproc.Semaphore == wrapper.Semaphore



"""Module containing tests for the different types of errors"""

# pylint: disable=missing-class-docstring

import pytest

from abllib import error

def test_customexception_inheritance():
    """Ensure that CustomException inherits from Exception"""

    assert issubclass(error.CustomException, Exception)

def test_customexception_notimplemented():
    """Ensure that direct CustomException usage raises errors"""

    with pytest.raises(NotImplementedError):
        raise error.CustomException()
    with pytest.raises(NotImplementedError):
        raise error.CustomException.with_values("test")

def test_customexception_custom_message():
    """Ensure that CustomException instantiation with a custom messaage keeps the message"""

    class TestException(error.CustomException):
        default_messages = {
            0: "This exception is used in tests"
        }

    try:
        raise TestException("This message should be retained")
    except TestException as e:
        assert str(e) == "This message should be retained"
    else:
        pytest.fail("Expected an exception")

def test_customexception_default_message():
    """Ensure that CustomException instantiation without a custom messaage uses the default message"""

    class TestException(error.CustomException):
        default_messages = {
            0: "This exception is used in tests"
        }

    try:
        raise TestException()
    except TestException as e:
        assert str(e) == "This exception is used in tests"
    else:
        pytest.fail("Expected an exception")

def test_wrongtypeerror():
    """Ensure that WrongTypeError handles optional args correctly"""

    try:
        raise error.WrongTypeError()
    except error.WrongTypeError as e:
        assert str(e) == "Received an unexpected type"
    else:
        pytest.fail("Expected an exception")

    try:
        raise error.WrongTypeError("This is a custom error message")
    except error.WrongTypeError as e:
        assert str(e) == "This is a custom error message"
    else:
        pytest.fail("Expected an exception")

    try:
        raise error.WrongTypeError.with_values("20", int)
    except error.WrongTypeError as e:
        assert str(e) == "Expected <class 'int'>, not <class 'str'>"
    else:
        pytest.fail("Expected an exception")

    try:
        raise error.WrongTypeError.with_values("20", 20)
    except error.WrongTypeError as e:
        assert str(e) == "Expected <class 'int'>, not <class 'str'>"
    else:
        pytest.fail("Expected an exception")

    try:
        raise error.WrongTypeError.with_values(str, int)
    except error.WrongTypeError as e:
        assert str(e) == "Expected <class 'int'>, not <class 'str'>"
    else:
        pytest.fail("Expected an exception")



"""A module containing custom errors"""

from ._custom_exception import CustomException
from ._general import CalledMultipleTimesError, \
                      DeprecatedError, \
                      DirNotFoundError, \
                      InternalCalculationError, \
                      InternalFunctionUsedError, \
                      InvalidKeyError, \
                      KeyNotFoundError, \
                      LockAcquisitionTimeoutError, \
                      MissingDefaultMessageError, \
                      MissingInheritanceError, \
                      NameNotFoundError, \
                      NoneTypeError, \
                      NotInitializedError, \
                      ReadonlyError, \
                      RegisteredMultipleTimesError, \
                      SingletonInstantiationError, \
                      UninitializedFieldError, \
                      WrongTypeError

INTERNAL =  "Internal error, please report it on github!"

__exports__ = [
    CustomException,
    CalledMultipleTimesError,
    DeprecatedError,
    DirNotFoundError,
    InternalCalculationError,
    InternalFunctionUsedError,
    InvalidKeyError,
    KeyNotFoundError,
    LockAcquisitionTimeoutError,
    MissingDefaultMessageError,
    MissingInheritanceError,
    NameNotFoundError,
    NoneTypeError,
    NotInitializedError,
    ReadonlyError,
    RegisteredMultipleTimesError,
    SingletonInstantiationError,
    UninitializedFieldError,
    WrongTypeError,
    INTERNAL
]



"""The main module, used for dev testing"""

# TODO:
# UP NEXT
# supply package number manually to release workflow
# publish to pypi
# replace all TypeError occurances with WrongTypeError

# IMPORTANT BUT ANNOYING
# add tests for threaded storage access
# document and understand levenshtein_distance
# document fuzzy score calculation

# TAKES A LOT OF TIME / NEEDED IN THE FUTURE
# add abllib.type module with verify function, which compares a value to a given type hint
# add a enforce_types wrapper for ensuring function parameter types, which internally uses abllib.type
# go throught all TODO comments and add type validation if necessary
# add async module with function to run async func in sync context

if __name__ == "__main__":
    pass



"""Module containing the _VolatileStorage class"""

from ._threadsafe_storage import _ThreadsafeStorage
from ..storage._storage_view import _StorageView
from .. import error

class _VolatileStorage(_ThreadsafeStorage):
    """Storage that is not saved across restarts"""

    def __init__(self) -> None:
        if _VolatileStorage._instance is not None:
            raise error.SingletonInstantiationError.with_values(_VolatileStorage)

        _VolatileStorage._instance = self

    def initialize(self):
        """
        Initialize only the VolatileStorage.

        Not needed if you already called abllib.storage.initialize().
        """

        if _VolatileStorage._store is not None:
            # this is a re-initialization
            return

        _VolatileStorage._store = self._store = {}

        # we cannot use StorageView defined in __init__.py because of circular imports
        # pylint: disable-next=protected-access
        _StorageView._instance.add_storage(self)

    _STORAGE_NAME = "VolatileStorage"

    def _ensure_initialized(self):
        try:
            super()._ensure_initialized()
        except error.NotInitializedError as exc:
            raise error.NotInitializedError("VolatileStorage is not yet initialized, "
                                            + "are you sure you called storage.initialize() "
                                            + "or VolatileStorage.initialize()?") \
                                           from exc



"""Module containing the _PersistentStorage class"""

from typing import Any

from .._storage._base_storage import _BaseStorage
from .. import error, wrapper

class _ThreadsafeStorage(_BaseStorage):
    def __init__(self):
        raise NotImplementedError()

    _STORAGE_NAME = "ThreadsafeStorage"

    @wrapper.NamedSemaphore(_STORAGE_NAME)
    def contains_item(self, key, item):
        return super().contains_item(key, item)

    @wrapper.NamedSemaphore(_STORAGE_NAME)
    def contains(self, key):
        return super().contains(key)

    @wrapper.NamedSemaphore(_STORAGE_NAME)
    def get(self, key, default = None):
        return super().get(key, default)

    @wrapper.NamedSemaphore(_STORAGE_NAME)
    def items(self):
        return super().items()

    @wrapper.NamedLock(_STORAGE_NAME)
    def pop(self, key) -> Any:
        return super().pop(key)

    @wrapper.NamedSemaphore(_STORAGE_NAME)
    def keys(self):
        return super().keys()

    @wrapper.NamedSemaphore(_STORAGE_NAME)
    def values(self):
        return super().values()

    @wrapper.NamedSemaphore(_STORAGE_NAME)
    def __getitem__(self, key):
        return super().__getitem__(key)

    @wrapper.NamedLock(_STORAGE_NAME)
    def __setitem__(self, key: str, item: Any) -> None:
        return super().__setitem__(key, item)

    @wrapper.NamedLock(_STORAGE_NAME)
    def __delitem__(self, key):
        return super().__delitem__(key)

    @wrapper.NamedSemaphore(_STORAGE_NAME)
    def __contains__(self, key):
        return super().__contains__(key)

    def __init_subclass__(cls):
        if cls._STORAGE_NAME in ("BaseStorage", "ThreadsafeStorage"):
            raise error.UninitializedFieldError.with_values(cls, "_STORAGE_NAME")

        if not isinstance(cls._STORAGE_NAME, str):
            raise error.WrongTypeError.with_values(cls._STORAGE_NAME, str)



"""Module containing the similarity function"""

import numpy as np

from abllib import error
from ..alg import levenshtein_distance

class Similarity():
    """
    Checks how closely two strings match. (Version 2)

    Returns a float value between 0.0 and 1.0 (inclusive), where 1.0 is a perfect match.
    """

    def __init__(self, target: str, candidate: str, threshold: int = 5) -> None:
        targets = target.split(" ")
        candidates = candidate.split(" ")

        # ensure that targets is always smaller than candidates
        if len(targets) > len(candidates):
            target, candidate = candidate, target
            targets, candidates = candidates, targets

        self._target = target
        self._candidate = candidate

        self._targets = targets
        self._candidates = candidates

        self._threshold = threshold

        self.scores_array = self._construct_scores_array()

    def _construct_scores_array(self) -> np.ndarray:
        """
        The returned 2D scores_array is structured as follows:
                    candidate1  candidate2  candidate3  candidate4
        target1     0.75        0.5         0.0         0.0
        target2     0.0         0.0         1.0         0.0
        target3     1.0         0.0         0.0         0.8
        such that scores[2][3] == 0.8.
        """

        scores_array = np.full((len(self._targets), len(self._candidates)), fill_value=0.0)

        # construct scores_array 2d array
        for i_target, inner_target in enumerate(self._targets):
            for i_candidate, inner_candidate in enumerate(self._candidates):
                edit_dist = levenshtein_distance(inner_target, inner_candidate)

                max_dist_by_target = (len(inner_target) // 3) + 1
                max_dist_by_candidate = (len(inner_candidate) // 3) + 1
                max_allowed_dist = min(max_dist_by_target, max_dist_by_candidate, self._threshold)

                # check if edit distance is within bounds
                if edit_dist <= max_allowed_dist:
                    # we take either inner_target or inner_candidate, depending on which has more characters
                    # this ensures that swapping target and candidate results in the same score
                    max_inner_len = max(len(inner_target), len(inner_candidate))
                    similar_chars = max_inner_len - edit_dist
                    divisor = max_inner_len
                    score = similar_chars / divisor
                    scores_array[i_target][i_candidate] = score

        return scores_array

    def calculate(self) -> float:
        """Calculate the similarity"""

        score = max(
            self._calculate_simple(),
            self._calculate_complex()
        )
        score = np.round(score, 2)

        if score < 0.0 or score > 1.0:
            raise error.InternalCalculationError(f"Score {score} is not in acceptable range 0.0 <= score <= 1.0")

        return score

    def _calculate_simple(self) -> float:
        edit_dist = levenshtein_distance(self._target, self._candidate)

        if edit_dist > self._threshold:
            return 0.0

        max_len = max(len(self._target), len(self._candidate))
        similar_chars = max_len - edit_dist
        divisor = max_len
        score = similar_chars / divisor
        return score

    def _calculate_complex(self) -> float:
        score_divisor = len(self._candidates)

        if not _contains_duplicates(self._construct_primitive_indexes()):
            # the target words don't interfere with each other, so we can use the best score for each

            total_score = 0.0
            for row in self.scores_array:
                total_score += row.max()
            return total_score / score_divisor

        indexes = self._construct_overlapping_indexes()

        optimal_indexes = self._construct_optimal_indexes(indexes)

        total_score = 0.0
        for row_i, col_i in enumerate(optimal_indexes):
            total_score += self.scores_array[row_i][col_i]

        return total_score / score_divisor

    def _construct_primitive_indexes(self) -> list[int]:
        indexes = []

        for row in self.scores_array:
            index = np.argmax(row)
            if row[index] != 0.0:
                indexes.append(index)

        return indexes

    def _construct_overlapping_indexes(self) -> dict[int, list[int]]:
        indexes = {}

        for row_i, row in enumerate(self.scores_array):
            index = np.argmax(row)
            if row[index] == 0.0:
                index = -1

            if index not in indexes:
                indexes[index] = []
            indexes[index].append(row_i)

        return indexes

    def _construct_optimal_indexes(self, indexes: dict[int, list[int]]) -> list[int]:
        """
        This class will create a list which maximizes the total score, so that each target is used once.
        For example:
                    index
        target1     0
        target2     2
        target3     3
        such that the total score equals 0.75 + 1.0 + 0.8 = 2.55.
        """

        while not self._optional_calc_is_done(indexes):
            curr_index = self._optional_calc_get_next_index(indexes)

            curr_target_indexes = indexes[curr_index]
            del indexes[curr_index]

            curr_scores = np.ndarray((len(curr_target_indexes), self.scores_array.shape[1]), dtype=np.ndarray)

            scores_i = 0
            for row_i, row in enumerate(self.scores_array):
                if row_i in curr_target_indexes:
                    curr_scores[scores_i] = row
                    scores_i += 1

            _, ind = _alg_with_index(curr_scores, 0.0)

            for target_i, candidate_i in zip(curr_target_indexes, ind):
                if candidate_i not in indexes:
                    indexes[candidate_i] = []
                indexes[candidate_i].append(target_i)

        number_of_targets = sum((len(item) for item in indexes.values()))
        optimal_indexes = np.full(number_of_targets, dtype=int, fill_value=-1)
        for candidate_i, target_i in indexes.items():
            if candidate_i != -1:
                optimal_indexes[target_i[0]] = candidate_i

        if -1 in indexes:
            # fill up remaining slots with target_i which didn't match (candidate_1 == -1)
            for target_i in indexes[-1]:
                for i, item in enumerate(optimal_indexes):
                    if item == -1:
                        optimal_indexes[i] = target_i

        return optimal_indexes

    def _optional_calc_is_done(self, indexes: dict[int, list[int]]) -> bool:
        for i, l in indexes.items():
            if i != -1 and len(l) > 1:
                return False
        return True

    def _optional_calc_get_next_index(self, indexes: dict[int, list[int]]) -> int:
        for i in indexes.keys():
            if i != -1 and len(indexes[i]) > 1:
                return i

        raise RuntimeError()

def _contains_duplicates(arr: np.ndarray) -> bool:
    seen = set()

    for item in arr:
        if item in seen:
            return True

        seen.add(item)

    return False

def _alg_with_index(data: np.ndarray, combined_score: float) -> tuple[float, list[int]]:
    """
    I have no idea what to call this algorithm.
    
    It generates all unique combinations of a 2d input array such that in each combination,
    each row and each column only occurs once.

    This process is done recursively and is quite costly,
    as the number of combinations equals !n, where n is (number of rows / number of columns).
    """

    if data.shape[0] == 1:
        # take the maximum score of all current scores
        index = int(data.argmax())
        if data[0][index] > 0.0:
            return (combined_score + data[0][index], [index])
        return (combined_score, [-1])

    max_score = 0.0
    max_indexes = None
    max_row_index = None

    for row_index in range(data.shape[0]):
        reduced_data = _reduce(data, row_index)

        score, indexes = _alg_with_index(reduced_data, combined_score + data[row_index][0])

        # we have to initialize at least once
        if max_indexes is None:
            max_indexes = indexes
            max_row_index = row_index

        # branching with if is much faster than max, because max_score only rarely changes
        # pylint: disable-next=consider-using-max-builtin
        if score > max_score:
            max_score = score
            max_indexes = indexes
            max_row_index = row_index

    # shift indexes
    for i, item in enumerate(max_indexes):
        if item != -1:
            max_indexes[i] = item + 1

    # store current index
    if data[max_row_index][0] != 0.0:
        max_indexes.insert(max_row_index, 0)
    else:
        max_indexes.insert(max_row_index, -1)

    return (max_score, max_indexes)

def _reduce(data: np.ndarray, r_index: int) -> np.ndarray:
    new_array = np.empty((data.shape[0] - 1, data.shape[1] - 1))

    row_i = 0
    for orig_row_i in range(data.shape[0]):
        if orig_row_i != r_index:
            # we can copy the whole row, excluding the first element
            new_array[row_i] = data[orig_row_i][1:]
            row_i += 1

    return new_array



"""Module containing the MatchResult dataclass"""

from dataclasses import dataclass

@dataclass
class MatchResult():
    """Dataclass representing a single matched candidate"""

    score: float
    value: str | None = None
    index: int | None = None
    inner_index: int | None = None



"""Module containing the _InternalStorage class"""

from typing import Any

from ._base_storage import _BaseStorage
from .. import error

class _InternalStorage(_BaseStorage):
    """Internal storage that is not saved across restarts"""

    def __init__(self) -> None:
        pass

    def _init(self):
        if _InternalStorage._instance is not None:
            raise error.SingletonInstantiationError.with_values(_InternalStorage)

        _InternalStorage._store = self._store = {}
        _InternalStorage._instance = self

    _STORAGE_NAME = "InternalStorage"

    def __setitem__(self, key: str, item: Any) -> None:
        if not key.startswith("_"):
            raise error.InternalFunctionUsedError("Please use storage.VolatileStorage")

        return super().__setitem__(key, item)



"""Module containing the _StorageView class"""

from __future__ import annotations
from typing import Any

from .._storage._base_storage import _BaseStorage
from ._cache_storage import _CacheStorage
from .. import error

# pylint: disable=protected-access

class _StorageView():
    """A read-only view on both the PersistentStorage and VolatileStorage"""

    def __init__(self):
        pass

    def _init(self) -> None:
        if _StorageView._instance is not None:
            raise error.SingletonInstantiationError.with_values(_StorageView)

        _StorageView._instance = self

        # CacheStorage needs to be added like this to avoid circular imports
        _StorageView._storages = self._storages = [_CacheStorage._instance]

    _instance: _StorageView = None
    _storages: list[_BaseStorage] = None

    def add_storage(self, storage: _BaseStorage) -> None:
        """
        Add a new storage to the StorageView.

        The storage has to inherit from _BaseStorage.

        If the exact same storage object is registered twice, a RegisteredMultipleTimesError error is raised.
        """

        if not isinstance(storage, _BaseStorage):
            raise error.MissingInheritanceError.with_values(storage, _BaseStorage)

        for item in self._storages:
            if id(item) == id(storage):
                raise error.RegisteredMultipleTimesError.with_values(storage)

        self._storages.append(storage)

    def contains_item(self, key: str, item: Any) -> bool:
        """
        Checks whether a key within the storage equals an item.

        If 'key' contains a '.', also checks if all sub-dicts exist.
        """

        for storage in self._storages:
            if storage.contains_item(key, item):
                return True
        return False

    def contains(self, key: str) -> bool:
        """
        Checks whether a key exists within the storage.

        If 'key' contains a '.', also checks if all sub-dicts exist.
        """

        for storage in self._storages:
            if storage.contains(key):
                return True
        return False

    def get(self, key: str, default: Any = None) -> Any:
        """
        Return the value of an key if it exists in the storage.

        If the key is not found, return the default value instead.
        """

        for storage in self._storages:
            if storage.contains(key):
                return storage[key]

        return default

    def items(self) -> list[tuple[str, Any]]:
        """
        Return the top-level keys and values in the storage.
        """

        items = []

        for storage in self._storages:
            items += list(storage.items())

        return items

    def keys(self):
        """
        Return the top-level keys in the storage.
        """

        keys = []

        for storage in self._storages:
            keys += list(storage.keys())

        return keys

    def values(self):
        """
        Return the top-level items in the storage.
        """

        values = []

        for storage in self._storages:
            values += list(storage.values())

        return values

    def __getitem__(self, key: str) -> Any:
        for storage in self._storages:
            if key in storage:
                return storage[key]
        raise error.KeyNotFoundError.with_values(key)

    def __contains__(self, key: str) -> bool:
        return self.contains(key)



"""
Ableytner's library for Python

Contains many general-purpose functions which can be used across projects.
"""

from . import alg, error, fs, fuzzy, onexit, wrapper
from .log import get_logger
from .storage import CacheStorage, VolatileStorage, PersistentStorage, StorageView

__exports__ = [
    alg,
    error,
    fs,
    fuzzy,
    onexit,
    wrapper,
    get_logger,
    CacheStorage,
    VolatileStorage,
    PersistentStorage,
    StorageView
]



"""A module containing fuzzy matching-related functionality"""

from ._all import match_all
from ._closest import match_closest
from ._matchresult import MatchResult
from ._similarity import Similarity

def similarity(target: str, candidate: str) -> float:
    """
    Checks how closely two strings match. (Version 2)

    Returns a float value between 0.0 and 1.0 (inclusive), where 1.0 is a perfect match.
    """

    return Similarity(target, candidate).calculate()

__exports__ = [
    match_all,
    match_closest,
    MatchResult,
    Similarity,
    similarity
]



"""Module containing the deprecated wrapper"""

import functools
import traceback
from typing import Callable

from .. import log
from ..error import DeprecatedError

class deprecated():
    """
    Mark a function or class as deprecated.

    Calling the function logs a deprecation warning.

    If the optional arg raise_exec is set, raises a deprecation error instead.
    """

    def __new__(cls, message: str | None | Callable = None, raise_exec: bool = False):
        if not raise_exec:
            return deprecated.warning(message)

        return deprecated.error(message)

    @staticmethod
    def warning(message: str | None | Callable = None):
        """
        Mark a function or class as deprecated.

        Calling the function logs a deprecation warning.
        """

        # deprecated.warning was called with a custom message
        if not callable(message):
            return _Deprecated(message, False)

        # deprecated.warning was not called, only used as a decorator
        return _Deprecated(None, False)(message)

    @staticmethod
    def error(message: str | None | Callable = None):
        """
        Mark a function or class as deprecated.

        Calling the function raises a DeprecatedError.
        """

        # deprecated.error was called with a custom message
        if not callable(message):
            return _Deprecated(message, True)

        # deprecated.error was not called, only used as a decorator
        return _Deprecated(None, True)(message)

class _Deprecated():
    def __init__(self, message: str | None, raise_exec: bool):
        self._message = message
        self._raise_exec = raise_exec

    _message: str | None
    _raise_exec: bool

    def __call__(self, func: Callable):
        """Called when the class instance is used as a decorator"""

        def wrapper(*args, **kwargs):
            """The wrapped function that is called on function execution"""

            # the default message is used
            if self._message is None:
                traces = traceback.format_list(traceback.extract_stack())
                traces.reverse()
                usage_line = traces[1].split("\n")[0].strip()
                message = f"The functionality '{func.__name__}' is deprecated but used here: {usage_line}"
            # the message was overwritten
            else:
                message = str(self._message)

            if self._raise_exec:
                raise DeprecatedError(message)

            log.get_logger("deprecated").warning(message)

            return func(*args, **kwargs)

        # https://stackoverflow.com/a/17705456/15436169
        functools.update_wrapper(wrapper, func)

        return wrapper



"""Module containing the pylint test"""

import os

import pytest

from abllib import log

logger = log.get_logger("pylint")

def test_pylint():
    """Checks if all git-tracked python files adhere to pylint rules"""

    if os.name == "nt":
        files = os.popen("git ls-files *.py").read()
        files = " ".join([file.strip() for file in files.split("\n")])

        pylint_output = []
        for line in os.popen(f"pylint {files}").readlines():
            if line.strip().strip("-") != "":
                pylint_output.append(line.strip())

        if not "Your code has been rated at 10.00/10" in pylint_output[-1]:
            for line in pylint_output:
                logger.warning(line)

            pytest.fail(pylint_output[-1])
    else:
        # logging pylint errors on linux doesn't work
        assert "Your code has been rated at 10.00/10" in os.popen("pylint $(git ls-files '*.py')").read()



"""Module containing the NamedLock and NamedSemaphore classes"""

from __future__ import annotations

import functools
import traceback
from datetime import datetime
from time import sleep

from ._lock import Lock, Semaphore
from ._deprecated import deprecated
from .. import error, log
from .._storage import InternalStorage

logger = log.get_logger("LockWrapper")

class _BaseNamedLock():
    """
    The base class for the NamedLock and NamedSemaphore classes.
    """

    def __init__(self, lock_name: str, timeout: int | float | None = None):
        if isinstance(timeout, int):
            timeout = float(timeout)

        # TODO: add type validation
        if not isinstance(lock_name, str):
            raise error.WrongTypeError.with_values(lock_name, str)
        if not isinstance(timeout, float) and timeout is not None:
            raise error.WrongTypeError.with_values(timeout, float, None)

        self._name = lock_name
        self._timeout = timeout

        if "_locks" not in InternalStorage:
            InternalStorage["_locks.global"] = Lock()
        self._allocation_lock: Lock = InternalStorage["_locks.global"]

    _name: str
    _lock: Lock | Semaphore
    _timeout: float | None
    _allocation_lock: Lock
    _other_lock: _BaseNamedLock | None = None

    @property
    def name(self) -> str:
        """Return the lock's name"""

        return self._name

    def acquire(self) -> None:
        """Acquire the lock, or throw an LockAcquisitionTimeoutError if timeout is not None"""

        if self._timeout is None:
            self._allocation_lock.acquire()

            # ensure the other lock is not held
            other = self._get_other()
            if other is not None:
                while other.locked():
                    sleep(0.025)

            if not self._lock.acquire():
                self._allocation_lock.release()
                raise error.LockAcquisitionTimeoutError()

            self._allocation_lock.release()
            return

        initial_time = datetime.now()
        if not self._allocation_lock.acquire(timeout=self._timeout):
            raise error.LockAcquisitionTimeoutError(error.INTERNAL)

        elapsed_time = (datetime.now() - initial_time).total_seconds()

        # ensure the other lock is not held
        other = self._get_other()
        if other is not None:
            while other.locked():
                sleep(0.025)
                elapsed_time += 0.025
                if elapsed_time > self._timeout:
                    self._allocation_lock.release()
                    raise error.LockAcquisitionTimeoutError()

        if not self._lock.acquire(timeout=self._timeout - elapsed_time):
            self._allocation_lock.release()
            raise error.LockAcquisitionTimeoutError()

        self._allocation_lock.release()

    def release(self) -> None:
        """Release the lock"""

        self._lock.release()

    def locked(self) -> bool:
        """Return whether the lock is currentyl held"""

        return self._lock.locked()

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def __call__(self, func):
        """Called when instance is used as a decorator"""

        def wrapper(*args, **kwargs):
            """The wrapped function that is called on function execution"""

            with self:
                ret = func(*args, **kwargs)

            return ret

        # https://stackoverflow.com/a/17705456/15436169
        functools.update_wrapper(wrapper, func)

        return wrapper

    def _get_other(self) -> _BaseNamedLock | None:
        raise NotImplementedError()

class NamedLock(_BaseNamedLock):
    """
    Make a function require a lock to be held during execution.

    Only a single NamedLock can hold the lock, but only if the NamedSemaphore is not currently held.

    Optionally provide a timeout in seconds,
    after which an LockAcquisitionTimeoutError is thrown (disabled if timeout is None).
    """

    def __init__(self, lock_name, timeout = None):
        super().__init__(lock_name, timeout)

        if f"_locks.{lock_name}.l" not in InternalStorage:
            InternalStorage[f"_locks.{lock_name}.l"] = Lock()

        self._lock = InternalStorage[f"_locks.{lock_name}.l"]

    def acquire(self):
        _log_callstack(f"NamedLock '{self.name}' was acquired here:")
        return super().acquire()

    def release(self):
        _log_callstack(f"NamedLock '{self.name}' was released here:")
        return super().release()

    def _get_other(self):
        if self._other_lock is not None:
            return self._other_lock

        if f"_locks.{self.name}.s" in InternalStorage:
            self._other_lock = InternalStorage[f"_locks.{self.name}.s"]
            return self._other_lock

        return None

class NamedSemaphore(_BaseNamedLock):
    """
    Make a function require a lock to be held during execution.

    Multiple NamedSemaphores can hold the same lock concurrently, but only if the NamedLock is not currently held.

    Optionally provide a timeout in seconds,
    after which an LockAcquisitionTimeoutError is thrown (disabled if timeout is None).
    """

    def __init__(self, lock_name, timeout = None):
        super().__init__(lock_name, timeout)

        if f"_locks.{lock_name}.s" not in InternalStorage:
            InternalStorage[f"_locks.{lock_name}.s"] = Semaphore(999)

        self._lock = InternalStorage[f"_locks.{lock_name}.s"]

    def acquire(self):
        _log_callstack(f"NamedSemaphore '{self.name}' was acquired here:")
        return super().acquire()

    def release(self):
        _log_callstack(f"NamedSemaphore '{self.name}' was released here:")
        return super().release()

    def _get_other(self):
        if self._other_lock is not None:
            return self._other_lock

        if f"_locks.{self.name}.l" in InternalStorage:
            self._other_lock = InternalStorage[f"_locks.{self.name}.l"]
            return self._other_lock

        return None

def _log_callstack(message: str):
    """Log the current callstack"""

    if log.get_loglevel() != log.LogLevel.ALL:
        return

    traces = traceback.format_list(traceback.extract_stack())
    traces.reverse()

    for line in traces:
        ignore = False
        for filename in ["_lock_wrapper.py", "_persistent_storage.py", "_volatile_storage.py", "_storage_view.py"]:
            if filename in line:
                ignore = True

        if not ignore:
            logger.debug(message + "\n" + line.strip())
            return

@deprecated
class WriteLock(NamedLock):
    """Deprecated alias for NamedLock"""

@deprecated
class ReadLock(NamedSemaphore):
    """Deprecated alias for NamedSemaphore"""



"""A module containing various wrappers"""

from ._lock import Lock, Semaphore
from ._lock_wrapper import NamedLock, NamedSemaphore, WriteLock, ReadLock
from ._singleuse_wrapper import singleuse
from ._deprecated import deprecated

__exports__ = [
    Lock,
    Semaphore,
    NamedLock,
    NamedSemaphore,
    WriteLock,
    ReadLock,
    singleuse,
    deprecated
]



"""A module containing json-like storages"""

from ._cache_storage import _CacheStorage
from ._persistent_storage import _PersistentStorage
from ._storage_view import _StorageView
from ._threadsafe_storage import _ThreadsafeStorage
from ._volatile_storage import _VolatileStorage

# pylint: disable=protected-access

def initialize(filename: str = "storage.json", save_on_exit: bool = False):
    """
    Initialize the storage module.

    If save_on_exit is set to True, automatically calls PersistentStorage.save_to_disk on application exit.
    """

    VolatileStorage.initialize()

    PersistentStorage.initialize(filename, save_on_exit)

CacheStorage = _CacheStorage()
PersistentStorage = _PersistentStorage()
VolatileStorage = _VolatileStorage()

StorageView = _StorageView()
StorageView._init()

__exports__ = [
    initialize,
    CacheStorage,
    PersistentStorage,
    StorageView,
    VolatileStorage,
    _CacheStorage,
    _PersistentStorage,
    _StorageView,
    _ThreadsafeStorage,
    _VolatileStorage
]



"""Module containing tests for the abllib.fuzzy module"""

# pylint: disable=protected-access

from abllib import fuzzy

def test_all():
    """Ensure that fuzzy.match_all works at all"""

    target = "fox"
    inputs = ["the slow white rat", "the quick brown fox", "different saying with many words"]

    results = fuzzy.match_all(target, inputs)
    assert len(results) == 1
    assert results[0].value == "the quick brown fox"
    assert results[0].index == 1
    assert results[0].inner_index is None

    target = "the"
    inputs = ["the slow white rat", "the quick brown fox", "different saying with many words"]

    results = fuzzy.match_all(target, inputs)
    assert len(results) == 2
    assert results[0].value == "the slow white rat"
    assert results[0].index == 0
    assert results[0].inner_index is None
    assert results[1].value == "the quick brown fox"
    assert results[1].index == 1
    assert results[1].inner_index is None

def test_all_fuzzy():
    """Ensure that fuzzy.match_all applies fuzzy logic"""

    target = "diferent"
    inputs = ["the slow white rat", "the quick brown fox", "different saying with many words"]

    # pylint: disable-next=use-implicit-booleaness-not-comparison
    assert len(fuzzy.match_all(target, inputs, 0)) == 0
    assert len(fuzzy.match_all(target, inputs, 1)) == 1
    assert len(fuzzy.match_all(target, inputs, 8)) == 1

def test_all_tuple():
    """Ensure that fuzzy.match_all handles tuple candidates correctly"""

    target = "diferent"
    inputs = [
        ("the slow white rat", "this sentence is diferent"),
        ("the quick brown fox", "something else"),
        "different saying with many words"
    ]

    m = fuzzy.match_all(target, inputs)

    # pylint: disable-next=use-implicit-booleaness-not-comparison
    assert len(m) == 2
    assert m[0].value == ("the slow white rat", "this sentence is diferent")
    assert m[1].value == "different saying with many words"

def test_all_matchresult():
    """Ensure that fuzzy.match_all asigns correct values to MatchResult"""

    target = "diferent"
    inputs = [
        ("the slow white rat", "this sentence is diferent"),
        ("the quick brown fox", "something else"),
        "different saying with many words"
    ]

    m = fuzzy.match_all(target, inputs)

    # pylint: disable-next=use-implicit-booleaness-not-comparison
    assert len(m) == 2
    assert m[0].index == 0
    assert m[0].value == ("the slow white rat", "this sentence is diferent")
    assert m[0].inner_index == 1

def test_closest():
    """Ensure that fuzzy.match_closest works at all"""

    target = "fox"
    inputs = ["the slow white rat", "the quick brown fox", "different saying with many words"]

    result = fuzzy.match_closest(target, inputs, 0)
    assert result.value == "the quick brown fox"
    assert result.index == 1
    assert result.inner_index is None

    target = "the"
    inputs = ["the slow white rat", "the quick brown fox", "different saying with many words"]

    result = fuzzy.match_closest(target, inputs, 0)
    assert result.value == "the slow white rat"
    assert result.index == 0
    assert result.inner_index is None

def test_closest_fuzzy():
    """Ensure that fuzzy.match_closest applies fuzzy logic"""

    target = "diferent"
    inputs = ["the slow white rat", "the quick brown fox", "different saying with many words"]

    assert fuzzy.match_closest(target, inputs, 0).value is None
    assert fuzzy.match_closest(target, inputs, 1).value == "different saying with many words"
    assert fuzzy.match_closest(target, inputs, 8).value == "different saying with many words"

    target = "diferent wth wors"
    inputs = ["the slow white rat", "the quick brown fox", "different saying with many words"]

    assert fuzzy.match_closest(target, inputs, 0).value is None
    assert fuzzy.match_closest(target, inputs, 1).value == "different saying with many words"
    assert fuzzy.match_closest(target, inputs, 8).value == "different saying with many words"

def test_closest_tuple():
    """Ensure that fuzzy.match_closest handles candidate tuples correctly"""

    target = "diferent"
    inputs = [
        ("the slow white rat", "this sentence is diferent"),
        ("the quick brown fox", "something else"),
        "different saying with many words"
    ]

    m = fuzzy.match_closest(target, inputs)

    assert m.value == ("the slow white rat", "this sentence is diferent")
    assert m.score == 0.25

def test_closest_matchresult():
    """Ensure that fuzzy.match_closest asigns correct values to MatchResult"""

    target = "diferent"
    inputs = [
        ("the slow white rat", "this sentence is diferent"),
        ("the quick brown fox", "something else"),
        "different saying with many words"
    ]

    m = fuzzy.match_closest(target, inputs)

    # pylint: disable-next=use-implicit-booleaness-not-comparison
    assert m.index == 0
    assert m.value == ("the slow white rat", "this sentence is diferent")
    assert m.inner_index == 1

def test_similarity():
    """Ensure that the similarity calculation works as expected"""

    similarity = fuzzy.similarity
    assert callable(similarity)

    assert similarity("fox", "the quick fox") == 0.33
    assert similarity("foy", "the quick fox") == 0.22

    # ensure single words also work
    assert similarity("fox", "fox") == 1
    assert similarity("dog", "dawg") == 0.5

    # only allow for an edit_distance of up to (len(word) // 3) + 1
    assert similarity("hoy", "the quick fox") == 0.11
    assert similarity("hay", "the quick fox") == 0.0

    assert similarity("sentence sentence",
                      "sentence candidate") == 0.5
    assert similarity("sentence sentence sentence",
                      "sentence candidate candidate") == 0.33
    assert similarity("fox",
                      "the quick fox") == 0.33
    assert similarity("sentence sen ntence",
                      "sentence sentence candidate") == 0.58
    assert similarity("this is a pretty pretty long target",
                      "a long pretty sentence is given as") == 0.57
    assert similarity("this is a pretty pretty long target sentence sentence sentence",
                     "a long pretty sentence is given as a candidate candidate") == 0.5
    assert similarity("first first second",
                      "first secon") == 0.61

def test_similarity_swappable():
    """Ensure that swapping the arguments for similarity calculation returns the same score"""

    similarity = fuzzy.similarity
    assert similarity("hoy", "the quick fox") == similarity("the quick fox", "hoy")
    assert similarity("sentence sen ntence", "sentence sentence candidate") \
           == similarity("sentence sentence candidate", "sentence sen ntence")



"""A module containing the fuzzy match function"""

from ._matchresult import MatchResult
from ._similarity import Similarity

def match_closest(target: str, candidates: list[str | tuple[str, ...]], threshold: int = 5) -> MatchResult:
    """
    Match the target to the most similar candidate. Applies fuzzy logic when comparing.

    In order to successfully match a candidate, at least one of two conditions need to be true:
    * the edit distance (levenshtein distance) needs to be smaller than *threshold*
    * a single word (*target* split at ' ') needs to have an edit distance smaller than (len(*word*) / 3) + 1

    After that, it chooses the closest-matching candidate.

    Returns a MatchResult
    """

    # TODO: type checking with abllib.type module

    if threshold < 0:
        raise ValueError("Threshold needs to be >= 0")

    result = MatchResult(0.0)
    for i, candidate in enumerate(candidates):
        curr_result = _matches_single_candidate(target, candidate, threshold)
        if curr_result.score > result.score:
            result = MatchResult(curr_result.score, curr_result.value, i, curr_result.inner_index)

    return result

def _matches_single_candidate(target: str, candidate: str | tuple[str, ...], threshold: int) -> MatchResult:
    if isinstance(candidate, str):
        score = Similarity(target, candidate, threshold).calculate()
        return MatchResult(score, candidate)

    result = MatchResult(0.0)
    for i, inner_candidate in enumerate(candidate):
        score = Similarity(target, inner_candidate, threshold).calculate()
        if score > result.score:
            result = MatchResult(score, candidate, inner_index=i)

    return result



"""Module containing tests for the different types of Storage"""

# pylint: disable=protected-access, missing-class-docstring, pointless-statement, expression-not-assigned

import json
import os

import pytest

from abllib import error, _storage
from abllib.storage import _CacheStorage, _VolatileStorage, _PersistentStorage, _StorageView, _ThreadsafeStorage
from abllib._storage._base_storage import _BaseStorage

def test_threadsafestorage_name_custom():
    """Ensure that custom storages need to overwrite _STORAGE_NAME"""

    ThreadsafeStorage = _ThreadsafeStorage.__new__(_ThreadsafeStorage)
    ThreadsafeStorage._store = {}

    with pytest.raises(error.UninitializedFieldError):
        class _TestStorage(_ThreadsafeStorage):
            def __init__(self):
                pass

    with pytest.raises(error.WrongTypeError):
        class _TestStorage2(_ThreadsafeStorage):
            def __init__(self):
                pass
            _STORAGE_NAME = 42

    with pytest.raises(error.UninitializedFieldError):
        class _TestStorage3(_ThreadsafeStorage):
            def __init__(self):
                pass
            _STORAGE_NAME = "BaseStorage"

    with pytest.raises(error.UninitializedFieldError):
        class _TestStorage4(_ThreadsafeStorage):
            def __init__(self):
                pass
            _STORAGE_NAME = "ThreadsafeStorage"

    class _TestStorage5(_ThreadsafeStorage):
        def __init__(self):
            pass
        _STORAGE_NAME = "TestStorage5"

def test_volatilestorage_inheritance():
    """Ensure the VolatileStorage inherits from _BaseStorage"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}

    assert isinstance(VolatileStorage, _BaseStorage)
    assert not isinstance(VolatileStorage, _PersistentStorage)
    assert not isinstance(VolatileStorage, _CacheStorage)

def test_volatilestorage_instantiation():
    """Ensure that VolatileStorage behaves like a singleton"""

    with pytest.raises(error.SingletonInstantiationError):
        _VolatileStorage()

    with pytest.raises(error.SingletonInstantiationError):
        _VolatileStorage()

def test_volatilestorage_valuetype():
    """Test the VolatileStorages' support for different value types"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}

    VolatileStorage["key1"] = ["1", 2, None]
    assert VolatileStorage["key1"] == ["1", 2, None]

    class CustomType():
        pass
    custom_item = CustomType()
    VolatileStorage["key1"] = custom_item
    assert VolatileStorage["key1"] == custom_item

def test_volatilestorage_noinit_error():
    """Ensure the VolatileStorage methods don't work before initialization is complete"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = None

    with pytest.raises(error.NotInitializedError):
        VolatileStorage["testkey"]
    with pytest.raises(error.NotInitializedError):
        VolatileStorage["testkey"] = "testvalue"
    with pytest.raises(error.NotInitializedError):
        del VolatileStorage["testkey"]
    with pytest.raises(error.NotInitializedError):
        VolatileStorage.contains("testkey")
    with pytest.raises(error.NotInitializedError):
        VolatileStorage.contains_item("testkey", "testvalue")

    try:
        VolatileStorage["testkey"]
    except error.NotInitializedError as exc:
        assert "VolatileStorage is not yet initialized" in str(exc)
    else:
        pytest.fail("expected exception")

def test_volatilestorage_del_autoremovedict():
    """Test that AutoremoveDicts are correctly deleted on del"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}

    VolatileStorage["key1.key2.key3.key4.key5.key6"] = "values"
    del VolatileStorage["key1.key2.key3.key4.key5.key6"]
    assert "key1.key2.key3.key4.key5" not in VolatileStorage
    assert "key1.key2.key3.key4" not in VolatileStorage
    assert "key1.key2.key3" not in VolatileStorage
    assert "key1.key2" not in VolatileStorage
    assert "key1" not in VolatileStorage

def test_persistentstorage_inheritance():
    """Ensure the PersistentStorage inherits from _BaseStorage"""

    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}

    assert isinstance(PersistentStorage, _BaseStorage)
    assert not isinstance(PersistentStorage, _VolatileStorage)
    assert not isinstance(PersistentStorage, _CacheStorage)

def test_persistentstorage_instantiation():
    """Ensure that PersistentStorage behaves like a singleton"""

    with pytest.raises(error.SingletonInstantiationError):
        _PersistentStorage()

    with pytest.raises(error.SingletonInstantiationError):
        _PersistentStorage()

def test_persistentstorage_valuetype():
    """Test the PersistentStorages' support for different value types"""

    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}

    PersistentStorage["key1"] = True
    assert PersistentStorage["key1"] is True
    PersistentStorage["key1"] = 10
    assert PersistentStorage["key1"] == 10
    PersistentStorage["key1"] = 10.1
    assert PersistentStorage["key1"] == 10.1
    PersistentStorage["key1"] = "value"
    assert PersistentStorage["key1"] == "value"
    PersistentStorage["key1"] = ["1", "2"]
    assert PersistentStorage["key1"] == ["1", "2"]
    PersistentStorage["key1"] = {"key": "item"}
    assert PersistentStorage["key1"] == {"key": "item"}
    PersistentStorage["key1"] = ("1", "2")
    assert PersistentStorage["key1"] == ("1", "2")
    PersistentStorage["key1"] = None
    assert PersistentStorage["key1"] is None

    class CustomType():
        pass

    with pytest.raises(TypeError):
        PersistentStorage["key1"] = CustomType()

def test_persistentstorage_load_file():
    """Test the PersistentStorage._load_from_disk() method"""

    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}

    filepath = _storage.InternalStorage["_storage_file"]

    with open(filepath, "w", encoding="utf8") as f:
        json.dump({
            "key1": "value",
            "key2": [
                "value21",
                "value22",
                "value23"
            ],
            "key3": 10
        }, f)

    PersistentStorage.load_from_disk()

    assert PersistentStorage["key1"] == "value"
    assert PersistentStorage["key2"] == ["value21", "value22", "value23"]
    assert PersistentStorage["key3"] == 10

    os.remove(filepath)

def test_persistentstorage_save_file():
    """Test the PersistentStorage._save_to_disk() method"""

    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}

    filepath = _storage.InternalStorage["_storage_file"]

    PersistentStorage["key1"] = "value"
    PersistentStorage["key2"] = ["value21", "value22", "value23"]
    PersistentStorage["key4"] = 10

    PersistentStorage.save_to_disk()

    assert os.path.isfile(filepath)
    with open(filepath, "r", encoding="utf8") as f:
        data = json.load(f)
    assert data["key1"] == "value"
    assert data["key2"] == ["value21", "value22", "value23"]
    assert data["key4"] == 10

    os.remove(filepath)

def test_persistentstorage_save_file_empty():
    """
    Ensure the PersistentStorage._save_to_disk() method doesn't overwrite an existing file
    if the PersistentStorage is empty
    """

    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}

    filepath = _storage.InternalStorage["_storage_file"]

    with open(filepath, "w", encoding="utf8") as f:
        json.dump({
            "key1": "newvalue"
        }, f)

    PersistentStorage.save_to_disk()

    with open(filepath, "r", encoding="utf8") as f:
        assert json.load(f)["key1"] == "newvalue"

def test_persistentstorage_noinit_error():
    """Ensure the PersistentStorage methods don't work before initialization is complete"""

    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = None

    with pytest.raises(error.NotInitializedError):
        PersistentStorage["testkey"]
    with pytest.raises(error.NotInitializedError):
        PersistentStorage["testkey"] = "testvalue"
    with pytest.raises(error.NotInitializedError):
        del PersistentStorage["testkey"]
    with pytest.raises(error.NotInitializedError):
        PersistentStorage.contains("testkey")
    with pytest.raises(error.NotInitializedError):
        PersistentStorage.contains_item("testkey", "testvalue")

    try:
        PersistentStorage["testkey"]
    except error.NotInitializedError as exc:
        assert "PersistentStorage is not yet initialized" in str(exc)
    else:
        pytest.fail("expected exception")

def test_persistentstorage_del_autoremovedict():
    """Test that AutoremoveDicts are correctly deleted on del"""

    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}

    PersistentStorage["key1.key2.key3.key4.key5.key6"] = "values"
    del PersistentStorage["key1.key2.key3.key4.key5.key6"]
    assert "key1.key2.key3.key4.key5" not in PersistentStorage
    assert "key1.key2.key3.key4" not in PersistentStorage
    assert "key1.key2.key3" not in PersistentStorage
    assert "key1.key2" not in PersistentStorage
    assert "key1" not in PersistentStorage

def test_storageview_instantiation():
    """Ensure that instantiating StorageView only works with valid arguments"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}
    CacheStorage = _CacheStorage.__new__(_CacheStorage)
    CacheStorage._store = {}
    StorageView = _StorageView()
    StorageView._storages = []
    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)
    StorageView.add_storage(CacheStorage)

    class FakeStorage():
        pass

    with pytest.raises(error.MissingInheritanceError):
        StorageView.add_storage(FakeStorage)

def test_storageview_getitem():
    """Test the Storage.__getitem__() method"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}
    StorageView = _StorageView()
    StorageView._storages = []
    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)

    VolatileStorage["key1"] = "value"
    PersistentStorage["key2"] = "value2"
    assert StorageView["key1"] == "value"
    assert StorageView["key2"] == "value2"

    del VolatileStorage["key1"]
    del PersistentStorage["key2"]

    VolatileStorage["key1.key2.key3"] = "value"
    PersistentStorage["key2.key3.key4.key5.key6.key7"] = "value2"
    assert StorageView["key1.key2.key3"] == "value"
    assert StorageView["key2.key3.key4.key5.key6.key7"] == "value2"

def test_storageview_keys():
    """Test the Storage.keys() method"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}
    StorageView = _StorageView()
    StorageView._storages = []
    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)

    assert len(StorageView.keys()) == 0
    assert "key1" not in StorageView.keys()
    assert "key2" not in StorageView.keys()

    VolatileStorage["key1"] = "value"
    PersistentStorage["key2"] = "value2"
    assert len(StorageView.keys()) == 2
    assert "key1" in StorageView.keys()
    assert "key2" in StorageView.keys()
    assert list(StorageView.keys())[0] in ["key1", "key2"]

def test_storageview_values():
    """Test the Storage.values() method"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}
    StorageView = _StorageView()
    StorageView._storages = []
    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)

    assert len(StorageView.values()) == 0
    assert "value" not in StorageView.values()
    assert "value2" not in StorageView.values()

    VolatileStorage["key1"] = "value"
    PersistentStorage["key2"] = "value2"
    assert len(StorageView.values()) == 2
    assert "value" in StorageView.values()
    assert "value2" in StorageView.values()
    assert list(StorageView.values())[0] in ["value", "value2"]

def test_storageview_items():
    """Test the Storage.items() method"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}
    StorageView = _StorageView()
    StorageView._storages = []
    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)

    assert len(StorageView.items()) == 0
    assert ("key1", "value") not in StorageView.items()
    assert ("key2", "value2") not in StorageView.items()

    VolatileStorage["key1"] = "value"
    PersistentStorage["key2"] = "value2"
    assert len(StorageView.items()) == 2
    assert ("key1", "value") in StorageView.items()
    assert ("key2", "value2") in StorageView.items()

def test_storageview_contains():
    """Test the Storage.contains() method"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}
    StorageView = _StorageView()
    StorageView._storages = []
    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)

    VolatileStorage["key1"] = "value"
    PersistentStorage["key2"] = "value2"
    assert StorageView.contains("key1")
    assert "key1" in StorageView
    assert StorageView.contains("key2")
    assert "key2" in StorageView

    del VolatileStorage["key1"]
    del PersistentStorage["key2"]

    VolatileStorage["key1.key2.key3"] = "value"
    PersistentStorage["key2.key3.key4.key5.key6.key7"] = "value2"
    assert StorageView.contains("key1.key2.key3")
    assert "key1.key2.key3" in StorageView
    assert StorageView.contains("key2.key3.key4.key5.key6.key7")
    assert "key2.key3.key4.key5.key6.key7" in StorageView

def test_storageview_contains_item():
    """Test the Storage.contains_item() method"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}
    StorageView = _StorageView()
    StorageView._storages = []
    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)

    VolatileStorage["key1"] = "value"
    PersistentStorage["key2"] = "value2"
    assert StorageView.contains_item("key1", "value")
    assert StorageView.contains_item("key2", "value2")

    del VolatileStorage["key1"]
    del PersistentStorage["key2"]

    VolatileStorage["key1.key2.key3"] = "value"
    PersistentStorage["key2.key3.key4.key5.key6.key7"] = "value2"
    assert StorageView.contains_item("key1.key2.key3", "value")
    assert StorageView.contains_item("key2.key3.key4.key5.key6.key7", "value2")

def test_storageview_get():
    """Test the Storage.get() method"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}
    StorageView = _StorageView()
    StorageView._storages = []
    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)

    assert not StorageView.contains("key1")
    assert StorageView.get("key1") is None

    VolatileStorage["key1"] = "value"
    PersistentStorage["key2"] = "value2"
    assert StorageView.contains("key1")
    assert StorageView.get("key1") == "value"
    assert StorageView.contains("key2")
    assert StorageView.get("key2") == "value2"

def test_storageview_uniqueness():
    """Test that StorageView.add_storage() only accepts each storage instance once"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}
    PersistentStorage2 = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage2._store = {}
    StorageView = _StorageView()
    StorageView._storages = []

    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)

    with pytest.raises(error.RegisteredMultipleTimesError):
        StorageView.add_storage(PersistentStorage)

    StorageView.add_storage(PersistentStorage2)

    with pytest.raises(error.RegisteredMultipleTimesError):
        StorageView.add_storage(PersistentStorage2)
    with pytest.raises(error.RegisteredMultipleTimesError):
        StorageView.add_storage(PersistentStorage2)

def test_storageview_get_default():
    """Test the Storage.get() methods' optional default argument"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = {}
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = {}
    StorageView = _StorageView()
    StorageView._storages = []
    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)

    assert not StorageView.contains("key1")
    assert StorageView.get("key1") is None

    assert StorageView.get("key1", None) is None
    assert StorageView.get("key1", default=None) is None

    assert StorageView.get("key1", 45) == 45
    assert StorageView.get("key1", default=45) == 45

    assert StorageView.get("key1", "test") == "test"
    assert StorageView.get("key1", default="test") == "test"

    VolatileStorage["key1"] = "value"
    PersistentStorage["key2"] = "value2"
    assert StorageView.contains("key1")
    assert StorageView.get("key1") == "value"
    assert StorageView.contains("key2")
    assert StorageView.get("key2") == "value2"

    assert StorageView.get("key1", None) == "value"
    assert StorageView.get("key1", default=None) == "value"
    assert StorageView.get("key2", None) == "value2"
    assert StorageView.get("key2", default=None) == "value2"

    assert StorageView.get("key1", 45) == "value"
    assert StorageView.get("key1", default=45) == "value"
    assert StorageView.get("key2", 45) == "value2"
    assert StorageView.get("key2", default=45) == "value2"

    assert StorageView.get("key1", "test") == "value"
    assert StorageView.get("key1", default="test") == "value"
    assert StorageView.get("key2", "test") == "value2"
    assert StorageView.get("key2", default="test") == "value2"

def test_storageview_noinit_error():
    """Ensure the StorageView methods don't work before initialization is complete"""

    VolatileStorage = _VolatileStorage.__new__(_VolatileStorage)
    VolatileStorage._store = None
    PersistentStorage = _PersistentStorage.__new__(_PersistentStorage)
    PersistentStorage._store = None
    StorageView = _StorageView()
    StorageView._storages = []
    StorageView.add_storage(VolatileStorage)
    StorageView.add_storage(PersistentStorage)

    with pytest.raises(error.NotInitializedError):
        StorageView["testkey"]
    with pytest.raises(error.NotInitializedError):
        StorageView.contains("testkey")
    with pytest.raises(error.NotInitializedError):
        StorageView.contains_item("testkey", "testvalue")

    try:
        StorageView["testkey"]
    except error.NotInitializedError as exc:
        # the first storage should raise an error
        assert "VolatileStorage is not yet initialized" in str(exc)
    else:
        pytest.fail("expected exception")

    VolatileStorage._store = {}

    try:
        StorageView["testkey"]
    except error.NotInitializedError as exc:
        # the second storage should raise an error
        assert "PersistentStorage is not yet initialized" in str(exc)
    else:
        pytest.fail("expected exception")

    PersistentStorage._store = {"testkey": "testval"}

    assert StorageView.contains_item("testkey", "testval")

def test_cachestorage_inheritance():
    """Ensure the CacheStorage inherits from _BaseStorage"""

    CacheStorage = _CacheStorage.__new__(_CacheStorage)
    CacheStorage._store = {}

    assert isinstance(CacheStorage, _BaseStorage)
    assert not isinstance(CacheStorage, _VolatileStorage)
    assert not isinstance(CacheStorage, _PersistentStorage)

def test_cachestorage_instantiation():
    """Ensure that CacheStorage behaves like a singleton"""

    with pytest.raises(error.SingletonInstantiationError):
        _CacheStorage()

    with pytest.raises(error.SingletonInstantiationError):
        _CacheStorage()

def test_cachestorage_valuetype():
    """Test the CacheStorages' support for different value types"""

    CacheStorage = _CacheStorage.__new__(_CacheStorage)
    CacheStorage._store = {}

    CacheStorage["key1"] = ["1", 2, None]
    assert CacheStorage["key1"] == ["1", 2, None]

    class CustomType():
        pass
    custom_item = CustomType()
    CacheStorage["key1"] = custom_item
    assert CacheStorage["key1"] == custom_item

def test_cachestorage_del_autoremovedict():
    """Test that AutoremoveDicts are correctly deleted on del"""

    CacheStorage = _CacheStorage.__new__(_CacheStorage)
    CacheStorage._store = {}

    CacheStorage["key1.key2.key3.key4.key5.key6"] = "values"
    del CacheStorage["key1.key2.key3.key4.key5.key6"]
    assert "key1.key2.key3.key4.key5" not in CacheStorage
    assert "key1.key2.key3.key4" not in CacheStorage
    assert "key1.key2.key3" not in CacheStorage
    assert "key1.key2" not in CacheStorage
    assert "key1" not in CacheStorage



"""Module containing the singleuse wrapper function"""

import functools

from ..error import CalledMultipleTimesError

def singleuse(func):
    """
    Make a function single-use only.
    If the function raised an exception, it is not seen as called and can be used again.

    Calling the function twice raises an error.CalledMultipleTimesError.
    """

    was_called = [False]

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        """The wrapped function that is called on function execution"""

        if was_called[0]:
            raise CalledMultipleTimesError()

        res = func(*args, **kwargs)

        was_called[0] = True

        return res

    return wrapper



"""Module containing tests for the abllib.onexit module"""

import pytest

from abllib import error, onexit

# pylint: disable=protected-access

def test_register():
    """Ensure that registering the same callback multiple times raises an error"""

    def func1():
        pass

    onexit.register("func1", func1)

    with pytest.raises(error.RegisteredMultipleTimesError):
        onexit.register("func1", func1)

    with pytest.raises(error.RegisteredMultipleTimesError):
        onexit.register("func1", func1)

def test_deregister():
    """Ensure that deregistering the same callback multiple times raises an error"""

    def func1():
        pass

    onexit.register("func1", func1)

    onexit.deregister("func1")

    with pytest.raises(error.NameNotFoundError):
        onexit.deregister("func1")

    with pytest.raises(error.NameNotFoundError):
        onexit.deregister("func1")

def test_register_single():
    """Ensure that registering the callbacks seperately works correctly"""

    def func1():
        pass

    onexit.register_normal_exit("func1", func1)

    onexit.deregister("func1")

    onexit.register_sigterm("func1", func1)

    onexit.deregister("func1")

    onexit.register_normal_exit("func1", func1)
    onexit.register_sigterm("func1", func1)

    onexit.deregister("func1")

def test_deregister_single():
    """Ensure that deregistering the callbacks seperately works correctly"""

    def func1():
        pass

    onexit.register("func1", func1)

    onexit.deregister_normal_exit("func1")

    onexit.register("func1", func1)

    onexit.deregister_sigterm("func1")

    onexit.register("func1", func1)

    onexit.deregister_normal_exit("func1")
    onexit.deregister_sigterm("func1")

def test_register_all():
    """Ensure that all register functions work together correctly"""

    def func1():
        pass

    onexit.register("func1", func1)

    onexit.deregister("func1")

    onexit.register("func1", func1)

def test_call_atexit():
    """Ensure that atexit function calls callbacks correctly"""

    data = [False]
    def func1():
        data[0] = True

    onexit.register("func1", func1)

    onexit._atexit_func()

    assert data[0]

def test_call_signal():
    """Ensure that signal function calls callbacks correctly"""

    data = [False]
    def func1():
        data[0] = True

    onexit.register("func1", func1)

    onexit._signal_func(None, None)

    assert data[0]

def test_dotname():
    """Ensure that names containing a "." work as expected"""

    data = [False]
    def func1():
        data[0] = True

    onexit.register("func1.cb", func1)

    onexit._atexit_func()

    assert data[0]



"""Module containing tests for the abllib.wrapper module"""

# pylint: disable=function-redefined, consider-using-with

from datetime import datetime

import pytest

from abllib import error, wrapper
from abllib.pproc import WorkerThread

def test_lock():
    """Ensure that Lock works as expected"""

    assert hasattr(wrapper, "Lock")
    assert callable(wrapper.Lock)

    lock = wrapper.Lock()

    assert not lock.locked()
    assert lock.acquire(blocking=True, timeout=1)
    assert lock.locked()

    assert not lock.acquire(blocking=True, timeout=1)

    lock.release()
    assert not lock.locked()

def test_semaphore():
    """Ensure that Semaphore works as expected"""

    assert hasattr(wrapper, "Semaphore")
    assert callable(wrapper.Semaphore)

    sem = wrapper.Semaphore(3)

    assert not sem.locked()
    assert sem.acquire(blocking=True, timeout=1)
    assert sem.locked()
    assert sem.acquire(blocking=True, timeout=1)
    assert sem.locked()
    assert sem.acquire(blocking=True, timeout=1)
    assert sem.locked()

    # the semaphore is full
    assert not sem.acquire(blocking=True, timeout=1)

    sem.release()
    assert sem.locked()
    sem.release()
    assert sem.locked()
    sem.release()
    assert not sem.locked()

def test_namedlock():
    """Ensure that NamedLock works as expected"""

    assert hasattr(wrapper, "NamedLock")
    assert callable(wrapper.NamedLock)

    @wrapper.NamedLock("test1", timeout=0.1)
    def func1():
        return True

    assert not wrapper.NamedLock("test1").locked()
    assert func1()
    assert not wrapper.NamedLock("test1").locked()

    wrapper.NamedLock("test2").acquire()
    assert wrapper.NamedLock("test2").locked()
    assert wrapper.NamedLock("test2", timeout=1).locked()

    wrapper.NamedLock("test3").acquire()
    def func2():
        assert wrapper.NamedLock("test3").locked()
        wrapper.NamedLock("test3", timeout=4).acquire()

    start_time = datetime.now()
    thread = WorkerThread(target=func2)
    thread.start()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        thread.join(reraise=True)

    duration = datetime.now() - start_time
    assert duration.total_seconds() > 3.5
    assert duration.total_seconds() < 4.5

def test_namedsemaphore():
    """Ensure that NamedSemaphore works as expected"""

    assert hasattr(wrapper, "NamedSemaphore")
    assert callable(wrapper.NamedSemaphore)

    @wrapper.NamedSemaphore("test1", timeout=0.1)
    def func1():
        return True

    assert not wrapper.NamedSemaphore("test1").locked()
    assert func1()
    assert not wrapper.NamedSemaphore("test1").locked()

    wrapper.NamedSemaphore("test2").acquire()
    assert wrapper.NamedSemaphore("test2").locked()
    assert wrapper.NamedSemaphore("test2", timeout=1).locked()

    wrapper.NamedLock("test3").acquire()
    def func2():
        assert not wrapper.NamedSemaphore("test3").locked()
        wrapper.NamedSemaphore("test3", timeout=4).acquire()

    start_time = datetime.now()
    thread = WorkerThread(target=func2)
    thread.start()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        thread.join(reraise=True)

    duration = datetime.now() - start_time
    assert duration.total_seconds() > 3.5
    assert duration.total_seconds() < 4.5

def test_namedlocks_combined():
    """Ensure that NamedLock and NamedSemaphore work together correctly"""

    @wrapper.NamedLock("test1", timeout=0.1)
    def func():
        return True

    wrapper.NamedSemaphore("test1").acquire()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        func()
    wrapper.NamedSemaphore("test1").release()


    @wrapper.NamedLock("test2", timeout=0.1)
    def func():
        return True

    wrapper.NamedSemaphore("test2").acquire()
    wrapper.NamedSemaphore("test2").acquire()
    wrapper.NamedSemaphore("test2").acquire()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        func()
    wrapper.NamedSemaphore("test2").release()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        func()
    wrapper.NamedSemaphore("test2").release()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        func()
    wrapper.NamedSemaphore("test2").release()
    func()


    @wrapper.NamedSemaphore("test3", timeout=0.1)
    def func():
        return True

    wrapper.NamedLock("test3").acquire()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        func()
    wrapper.NamedLock("test3").release()

def test_locks_underscore_names():
    """Ensure that named lock names can start with an underscore"""

    lock = wrapper.NamedSemaphore("_test1")
    assert not lock.locked()
    lock.acquire()
    assert lock.locked()
    assert wrapper.NamedSemaphore("_test1").locked()
    lock.release()
    assert not lock.locked()
    assert not wrapper.NamedSemaphore("_test1").locked()

    lock = wrapper.NamedLock("_test2")
    assert not lock.locked()
    lock.acquire()
    assert lock.locked()
    assert wrapper.NamedLock("_test2").locked()
    lock.release()
    assert not lock.locked()
    assert not wrapper.NamedLock("_test2").locked()

def test_singleuse():
    """Ensure that singleuse works as expected"""

    @wrapper.singleuse
    def func1():
        pass

    func1()

    with pytest.raises(error.CalledMultipleTimesError):
        func1()

    with pytest.raises(error.CalledMultipleTimesError):
        func1()

def test_singleuse_exception():
    """Ensure that raised exceptions reset singleuse flag"""

    data = [1, 2, 3]

    @wrapper.singleuse
    def func1():
        if len(data) > 0:
            data.pop(0)
            raise error.InternalCalculationError()

    with pytest.raises(error.InternalCalculationError):
        func1()
    with pytest.raises(error.InternalCalculationError):
        func1()
    with pytest.raises(error.InternalCalculationError):
        func1()

    func1()

    with pytest.raises(error.CalledMultipleTimesError):
        func1()

    with pytest.raises(error.CalledMultipleTimesError):
        func1()



"""An internal module containing json-like storages"""

# pylint: disable=protected-access

from ._base_storage import _BaseStorage
from ._internal_storage import _InternalStorage

InternalStorage = _InternalStorage()
# pylint: disable-next=protected-access
InternalStorage._init()

__exports__ = [
    _BaseStorage,
    InternalStorage
]



"""A module containing parallel processing-related functionality, both with threads and processes"""

from ._worker_process import WorkerProcess
from ._worker_thread import WorkerThread
from ..wrapper import Lock, Semaphore

__exports__ = [
    WorkerProcess,
    WorkerThread,
    Lock,
    Semaphore
]



"""Module containing tests for the abllib.alg module"""

import pytest

from abllib import alg, error

def test_levenshtein_distance():
    """Ensure that alg.levenshtein_distance works as expected"""

    assert alg.levenshtein_distance("fox", "fox") == 0
    assert alg.levenshtein_distance("fox", "fof") == 1
    assert alg.levenshtein_distance("the brown fox", "the green fox") == 3
    assert alg.levenshtein_distance("", "a") == 1

    assert isinstance(alg.levenshtein_distance("dog", "god"), int)

    with pytest.raises(error.WrongTypeError):
        alg.levenshtein_distance("test", None)
    with pytest.raises(error.WrongTypeError):
        alg.levenshtein_distance(None, "test")
    with pytest.raises(error.WrongTypeError):
        alg.levenshtein_distance("test", 12)



"""A module containing the fuzzy search function"""

from ._matchresult import MatchResult
from ._similarity import Similarity

def match_all(target: str, candidates: list[str | tuple[str, ...]], threshold: int = 5) -> list[MatchResult]:
    """
    Search for all candidates matching the target. Applies fuzzy logic when comparing.

    In order to successfully find a candidate, at least one of two conditions need to be true:
    * the edit distance (levenshtein distance) needs to be smaller than *threshold*
    * a single word (*target* split at ' ') needs to have an edit distance smaller than (len(*word*) / 3) + 1

    Returns a list of MatchResults.
    """

    # TODO: type checking with abllib.type module

    if threshold < 0:
        raise ValueError("Threshold needs to be >= 0")

    results = []
    for i, candidate in enumerate(candidates):
        curr_result = _match_single_candidate(target, candidate, threshold)
        if curr_result is not None:
            results.append(MatchResult(curr_result.score, curr_result.value, i, curr_result.inner_index))

    return results

def _match_single_candidate(target: str, candidate: str | tuple[str, ...], threshold: int) -> MatchResult | None:
    if isinstance(candidate, str):
        score = Similarity(target, candidate, threshold).calculate()
        return MatchResult(score, candidate) if score > 0.0 else None

    result = MatchResult(0.0)
    for i, inner_candidate in enumerate(candidate):
        score = Similarity(target, inner_candidate, threshold).calculate()
        if score > result.score:
            result = MatchResult(score, candidate, inner_index=i)

    return result if result.score > 0.0 else None



"""
    Pytest fixtures
"""

# pylint: disable=protected-access, missing-class-docstring

import os
import shutil

import pytest

from abllib import fs, log, storage, _storage, onexit

logger = log.get_logger("test")

@pytest.fixture(scope="session", autouse=True)
def setup():
    """Setup the PersistentStorage, VolatileStorage and StorageView for test usage"""

    # setup testing dirs
    STORAGE_DIR = fs.absolute(os.path.dirname(__file__), "..", "..", "test_run")
    shutil.rmtree(STORAGE_DIR, ignore_errors=True)
    os.makedirs(STORAGE_DIR, exist_ok=True)

    #  setup logging
    log.initialize(log.LogLevel.DEBUG)
    log.add_console_handler()
    log.add_file_handler(os.path.join(STORAGE_DIR, "test.log"))

    STORAGE_FILE = fs.absolute(STORAGE_DIR, "test.json")

    if os.path.isfile(STORAGE_FILE):
        os.remove(STORAGE_FILE)

    storage.initialize(STORAGE_FILE)

    yield None

    storage.PersistentStorage.save_to_disk()

@pytest.fixture(scope="function", autouse=True)
def clean_after_function():
    """Clean up the PersistentStorage, VolatileStorage and StorageView, removing all keys"""

    yield None

    for key in list(storage.PersistentStorage._store.keys()):
        del storage.PersistentStorage[key]

    for key in list(storage.VolatileStorage._store.keys()):
        del storage.VolatileStorage[key]

    for key in list(_storage.InternalStorage._store.keys()):
        if key not in ["_storage_file", "_onexit"]:
            del _storage.InternalStorage[key]

    onexit.reset()

@pytest.fixture(scope="function", autouse=False)
def capture_logs():
    """Save all log output to a new file test.log in the root dir"""

    log.initialize(log.LogLevel.DEBUG)
    log.add_file_handler("test.log")

    yield None

    log.initialize()
    os.remove("test.log")



"""Module containing tests for the abllib.fs module"""

# pylint: disable=missing-class-docstring

import pathlib
import os

import pytest

from abllib import fs

def test_absolute():
    """Ensure that fs.absolute works as expected"""

    assert callable(fs.absolute)
    assert fs.absolute(os.getcwd()) == os.getcwd().replace("c:\\", "C:\\")
    assert fs.absolute("test.txt") == os.path.join(_uppercase_path(os.getcwd()), "test.txt")
    assert fs.absolute("subdir", "another", "test.txt") \
           == os.path.join(_uppercase_path(os.getcwd()), "subdir", "another", "test.txt")
    assert fs.absolute("subdir", pathlib.Path("another"), "test.txt") \
           == os.path.join(_uppercase_path(os.getcwd()), "subdir", "another", "test.txt")
    assert fs.absolute("subdir", "..", "test.txt") == os.path.join(_uppercase_path(os.getcwd()), "test.txt")

    with pytest.raises(TypeError):
        fs.absolute(None)
    with pytest.raises(TypeError):
        fs.absolute(1)
    with pytest.raises(TypeError):
        fs.absolute("one", "two", 3)
    with pytest.raises(ValueError):
        fs.absolute()

def _uppercase_path(path: str) -> str:
    if path.startswith("/"):
        return path

    first = path[0]
    first = first.upper()
    return first + path[1:]



"""Module containing the _PersistentStorage class"""

# pylint: disable=protected-access

import json
import os
from typing import Any

from .._storage import InternalStorage
from ._threadsafe_storage import _ThreadsafeStorage
from ..storage._storage_view import _StorageView
from .. import error, fs, onexit

class _PersistentStorage(_ThreadsafeStorage):
    """Storage that persists across restarts"""

    def __init__(self) -> None:
        if _PersistentStorage._instance is not None:
            raise error.SingletonInstantiationError.with_values(_PersistentStorage)

        _PersistentStorage._instance = self

    def initialize(self, filename: str = "storage.json", save_on_exit: bool = False):
        """
        Initialize only the PersistentStorage.

        Not needed if you already called abllib.storage.initialize().

        If save_on_exit is set to True, automatically calls save_to_disk on application exit.
        """

        full_filepath = fs.absolute(filename)
        if not os.path.isdir(os.path.dirname(full_filepath)):
            raise error.DirNotFoundError.with_values(os.path.dirname(full_filepath))

        if _PersistentStorage._store is not None:
            # this is a re-initialization
            if save_on_exit:
                try:
                    onexit.register("PersistentStorage.save", self.save_to_disk)
                except error.RegisteredMultipleTimesError:
                    pass
            else:
                try:
                    onexit.deregister("PersistentStorage.save")
                except error.NameNotFoundError:
                    pass

            if InternalStorage.contains_item("_storage_file", full_filepath):
                # the storage file didn't change
                pass
            else:
                # the storage file changed
                # save current store to old file
                self.save_to_disk()

                InternalStorage["_storage_file"] = full_filepath
                self.load_from_disk()

            return

        _PersistentStorage._store = self._store = {}

        _StorageView._instance.add_storage(self)

        InternalStorage["_storage_file"] = full_filepath
        self.load_from_disk()

        if save_on_exit:
            onexit.register("PersistentStorage.save", self.save_to_disk)

    _STORAGE_NAME = "PersistentStorage"

    def __setitem__(self, key: str, item: Any) -> None:
        # TODO: type check list / dict content types

        if not isinstance(item, (bool, int, float, str, list, dict, tuple)) and item is not None:
            raise TypeError(f"Tried to add item with type {type(item)} to PersistentStorage")

        return super().__setitem__(key, item)

    def load_from_disk(self) -> None:
        """Load the data from the storage file"""

        if "_storage_file" not in InternalStorage:
            raise error.KeyNotFoundError()

        path = InternalStorage["_storage_file"]
        if not os.path.isfile(path):
            return

        with open(path, "r", encoding="utf8") as f:
            self._store = json.load(f)

    def save_to_disk(self) -> None:
        """Save the data to the storage file"""

        if "_storage_file" not in InternalStorage:
            raise error.KeyNotFoundError()

        path = InternalStorage["_storage_file"]
        if len(self._store) == 0 and os.path.isfile(path):
            return

        with open(path, "w", encoding="utf8") as f:
            json.dump(self._store, f)

    def _ensure_initialized(self):
        try:
            super()._ensure_initialized()
        except error.NotInitializedError as exc:
            raise error.NotInitializedError("PersistentStorage is not yet initialized, "
                                            + "are you sure you called storage.initialize()?") \
                                           from exc



"""Module containing tests for the different types of Storage"""

# pylint: disable=protected-access, missing-class-docstring, pointless-statement, expression-not-assigned

import pytest

from abllib.error import InternalFunctionUsedError, \
                         InvalidKeyError, \
                         KeyNotFoundError, \
                         ReadonlyError, \
                         SingletonInstantiationError, \
                         UninitializedFieldError, \
                         WrongTypeError
from abllib._storage import _BaseStorage, _InternalStorage

def test_basestorage_instantiation():
    """Ensure that BaseStorage cannot be initialized"""

    with pytest.raises(NotImplementedError):
        _BaseStorage()

def test_basestorage_getitem():
    """Test the Storage.__getitem__() method"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage._store["key1"] = "value"
    assert BaseStorage["key1"] == "value"

    BaseStorage._store["key1"] = {}
    BaseStorage._store["key1"]["key2"] = "value2"
    assert BaseStorage["key1"]["key2"] == "value2"

def test_basestorage_getitem_multi():
    """Test the Storage.__getitem__() method with subdicts specified in the key"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage._store["key1"] = {}
    BaseStorage._store["key1"]["key2"] = "value2"
    assert BaseStorage["key1.key2"] == "value2"

    BaseStorage._store["key1"] = {}
    BaseStorage._store["key1"]["key2"] = {}
    BaseStorage._store["key1"]["key2"]["key3"] = {}
    BaseStorage._store["key1"]["key2"]["key3"]["key4"] = {}
    BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"] = {}
    BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"]["key6"] = "values"
    assert BaseStorage["key1.key2.key3.key4.key5.key6"] == "values"

def test_basestorage_getitem_keytype():
    """Test the Storage.__getitem__() methods' protection against incorrect key types"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(WrongTypeError):
        BaseStorage[None]
    with pytest.raises(WrongTypeError):
        BaseStorage[10]
    with pytest.raises(WrongTypeError):
        BaseStorage[list(("1",))]

    with pytest.raises(InvalidKeyError):
        BaseStorage[".some.key"]
    with pytest.raises(InvalidKeyError):
        BaseStorage["some.key."]
    with pytest.raises(InvalidKeyError):
        BaseStorage["some..key"]

def test_basestorage_getitem_valuetype():
    """Test the Storage.__getitem__() methods' support for different value types"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage._store["key1"] = ["1", 2, None]
    assert BaseStorage["key1"] == ["1", 2, None]

def test_basestorage_getitem_wrong_key():
    """Test the Storage.__getitem__() methods' protection against nonexistent keys"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(KeyNotFoundError):
        BaseStorage["key1"]
    with pytest.raises(KeyNotFoundError):
        BaseStorage["key1.key2"]
    with pytest.raises(KeyNotFoundError):
        BaseStorage["key1.key2.key3.key4.key5.key6"]

def test_basestorage_setitem():
    """Test the Storage.__setitem__() method"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage["key1"] = "value"
    assert BaseStorage._store["key1"] == "value"

    BaseStorage["key1"] = {}
    BaseStorage["key1"]["key2"] = "value2"
    assert BaseStorage._store["key1"]["key2"] == "value2"

def test_basestorage_setitem_multi():
    """Test the Storage.__setitem__() method with subdicts specified in the key"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage["key1"] = {}
    BaseStorage["key1.key2"] = "value2"
    assert BaseStorage._store["key1"]["key2"] == "value2"

    BaseStorage["key1"] = {}
    BaseStorage["key1.key2"] = {}
    BaseStorage["key1.key2.key3"] = {}
    BaseStorage["key1.key2.key3.key4"] = {}
    BaseStorage["key1.key2.key3.key4.key5"] = {}
    BaseStorage["key1.key2.key3.key4.key5.key6"] = "values"
    assert BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"]["key6"] == "values"

def test_basestorage_setitem_create_subdict():
    """Test the Storage.__setitem__() methods' ability to create missing 'inbetween' dictionaries"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage["key1.key2"] = "value2"
    assert isinstance(BaseStorage._store["key1"], dict)
    assert BaseStorage._store["key1"]["key2"] == "value2"

    BaseStorage["key1.key2"] = {}

    BaseStorage["key1.key2.key3.key4.key5.key6"] = "values"
    assert isinstance(BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"], dict)
    assert BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"]["key6"] == "values"

def test_basestorage_setitem_keytype():
    """Test the Storage.__setitem__() methods' protection against incorrect key types"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(WrongTypeError):
        BaseStorage[None] = "value"
    with pytest.raises(WrongTypeError):
        BaseStorage[10] = "value"
    with pytest.raises(WrongTypeError):
        BaseStorage[list(("1",))] = "value"

    with pytest.raises(InvalidKeyError):
        BaseStorage[".some.key"] = "value"
    with pytest.raises(InvalidKeyError):
        BaseStorage["some.key."] = "value"
    with pytest.raises(InvalidKeyError):
        BaseStorage["some..key"] = "value"

def test_basestorage_setitem_valuetype():
    """Test the Storage.__setitem__() methods' support for different value types"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage["key1"] = ["1", 2, None]
    assert BaseStorage._store["key1"] == ["1", 2, None]

    class CustomType():
        pass
    custom_item = CustomType()
    BaseStorage["key1"] = custom_item
    assert BaseStorage._store["key1"] == custom_item

def test_basestorage_delitem():
    """Test the Storage.__delitem__() method"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage._store["key1"] = "value"
    del BaseStorage["key1"]
    assert "key1" not in BaseStorage._store

def test_basestorage_delitem_multi():
    """Test the Storage.__delitem__() method with subdicts specified in the key"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage._store["key1"] = {}
    BaseStorage._store["key1"]["key2"] = "value2"
    del BaseStorage["key1.key2"]
    assert isinstance(BaseStorage._store["key1"], dict)
    assert "key2" not in BaseStorage._store["key1"]

    BaseStorage._store = {}

    BaseStorage._store["key1"] = {}
    BaseStorage._store["key1"]["key2"] = {}
    BaseStorage._store["key1"]["key2"]["key3"] = {}
    BaseStorage._store["key1"]["key2"]["key3"]["key4"] = {}
    BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"] = {}
    BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"]["key6"] = "values"
    del BaseStorage["key1.key2.key3.key4.key5.key6"]
    assert isinstance(BaseStorage._store["key1"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"]["key3"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"]["key3"]["key4"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"], dict)
    assert "key6" not in BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"]

    BaseStorage._store = {}

    BaseStorage["key1.key2.key3.key4.key5.key6"] = "values"
    BaseStorage["key1.key2.key3.key4.key5.another"] = "value2"
    del BaseStorage["key1.key2.key3.key4.key5.key6"]
    assert isinstance(BaseStorage._store["key1"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"]["key3"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"]["key3"]["key4"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"], dict)
    assert "another" in BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"]
    assert "key6" not in BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"]

    BaseStorage._store = {}

    BaseStorage["key1.key2.key3.key4.key5.key6"] = "values"
    del BaseStorage["key1.key2.key3.key4.key5.key6"]
    assert "key1.key2.key3.key4.key5" not in BaseStorage
    assert "key1.key2.key3.key4" not in BaseStorage
    assert "key1.key2.key3" not in BaseStorage
    assert "key1.key2" not in BaseStorage
    assert "key1" not in BaseStorage

def test_basestorage_delitem_keytype():
    """Test the Storage.__delitem__() methods' protection against incorrect key types"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(WrongTypeError):
        del BaseStorage[None]
    with pytest.raises(WrongTypeError):
        del BaseStorage[10]
    with pytest.raises(WrongTypeError):
        del BaseStorage[list(("1",))]

    with pytest.raises(InvalidKeyError):
        del BaseStorage[".some.key"]
    with pytest.raises(InvalidKeyError):
        del BaseStorage["some.key."]
    with pytest.raises(InvalidKeyError):
        del BaseStorage["some..key"]

def test_basestorage_delitem_wrong_key():
    """Test the Storage.__delitem__() methods' protection against nonexistent keys"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(KeyNotFoundError):
        del BaseStorage["key1"]
    with pytest.raises(KeyNotFoundError):
        del BaseStorage["key1.key2"]
    with pytest.raises(KeyNotFoundError):
        del BaseStorage["key1.key2.key3.key4.key5.key6"]

def test_basestorage_keys():
    """Test the Storage.keys() method"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert len(BaseStorage.keys()) == 0
    assert "key1" not in BaseStorage.keys()

    BaseStorage["key1"] = "value"
    assert len(BaseStorage.keys()) == 1
    assert "key1" in BaseStorage.keys()
    assert list(BaseStorage.keys())[0] == "key1"

def test_basestorage_values():
    """Test the Storage.values() method"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert len(BaseStorage.values()) == 0
    assert "value" not in BaseStorage.values()

    BaseStorage["key1"] = "value"
    assert len(BaseStorage.values()) == 1
    assert "value" in BaseStorage.values()
    assert list(BaseStorage.values())[0] == "value"

def test_basestorage_items():
    """Test the Storage.items() method"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert len(BaseStorage.items()) == 0
    assert ("key1", "value") not in BaseStorage.items()

    BaseStorage["key1"] = "value"
    assert len(BaseStorage.items()) == 1
    assert ("key1", "value") in BaseStorage.items()

def test_basestorage_pop():
    """Test the Storage.pop() method"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage._store["key1"] = "value"
    val = BaseStorage.pop("key1")
    assert val == "value"
    assert "key1" not in BaseStorage._store

def test_basestorage_pop_multi():
    """Test the Storage.pop() method with subdicts specified in the key"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    BaseStorage._store["key1"] = {}
    BaseStorage._store["key1"]["key2"] = "value2"
    val = BaseStorage.pop("key1.key2")
    assert val == "value2"
    assert isinstance(BaseStorage._store["key1"], dict)
    assert "key2" not in BaseStorage._store["key1"]

    BaseStorage._store["key1"] = {}
    BaseStorage._store["key1"]["key2"] = {}
    BaseStorage._store["key1"]["key2"]["key3"] = {}
    BaseStorage._store["key1"]["key2"]["key3"]["key4"] = {}
    BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"] = {}
    BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"]["key6"] = "values"
    val = BaseStorage.pop("key1.key2.key3.key4.key5.key6")
    assert val == "values"
    assert isinstance(BaseStorage._store["key1"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"]["key3"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"]["key3"]["key4"], dict)
    assert isinstance(BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"], dict)
    assert "key2" not in BaseStorage._store["key1"]["key2"]["key3"]["key4"]["key5"]

def test_basestorage_pop_keytype():
    """Test the Storage.pop() methods' protection against incorrect key types"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(WrongTypeError):
        BaseStorage.pop(None)
    with pytest.raises(WrongTypeError):
        BaseStorage.pop(10)
    with pytest.raises(WrongTypeError):
        BaseStorage.pop(list(("1",)))

    with pytest.raises(InvalidKeyError):
        BaseStorage.pop(".some.key")
    with pytest.raises(InvalidKeyError):
        BaseStorage.pop("some.key.")
    with pytest.raises(InvalidKeyError):
        BaseStorage.pop("some..key")

def test_basestorage_pop_wrong_key():
    """Test the Storage.pop() methods' protection against nonexistent keys"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(KeyNotFoundError):
        BaseStorage.pop("key1")
    with pytest.raises(KeyNotFoundError):
        BaseStorage.pop("key1.key2")
    with pytest.raises(KeyNotFoundError):
        BaseStorage.pop("key1.key2.key3.key4.key5.key6")

def test_basestorage_contains():
    """Test the Storage.contains() method"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert not BaseStorage.contains("key1")
    assert "key1" not in BaseStorage

    BaseStorage["key1"] = "value"
    assert BaseStorage.contains("key1")
    assert "key1" in BaseStorage

def test_basestorage_contains_multi():
    """Test the Storage.contains() method with subdicts specified in the key"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert not BaseStorage.contains("key1.key2")
    assert "key1.key2" not in BaseStorage

    BaseStorage["key1.key2"] = "value2"
    assert BaseStorage.contains("key1.key2")
    assert "key1.key2" in BaseStorage

    del BaseStorage["key1"]

    assert not BaseStorage.contains("key1.key2.key3.key4.key5.key6")
    assert "key1.key2.key3.key4.key5.key6" not in BaseStorage

    BaseStorage["key1.key2.key3.key4.key5.key6"] = "values"
    assert BaseStorage.contains("key1.key2.key3.key4.key5.key6")
    assert "key1.key2.key3.key4.key5.key6" in BaseStorage

def test_basestorage_contains_keytype():
    """Test the Storage.contains() methods' protection against incorrect key types"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(WrongTypeError):
        None in BaseStorage
    with pytest.raises(WrongTypeError):
        10 in BaseStorage
    with pytest.raises(WrongTypeError):
        list(("1",)) in BaseStorage

    with pytest.raises(InvalidKeyError):
        ".some.key" in BaseStorage
    with pytest.raises(InvalidKeyError):
        "some.key." in BaseStorage
    with pytest.raises(InvalidKeyError):
        "some..key" in BaseStorage

def test_basestorage_contains_item():
    """Test the Storage.contains_item() method"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert not BaseStorage.contains_item("key1", "value")

    BaseStorage["key1"] = "value"
    assert BaseStorage.contains_item("key1", "value")

def test_basestorage_contains_item_multi():
    """Test the Storage.contains_item() method with subdicts specified in the key"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert not BaseStorage.contains_item("key1.key2", "value")

    BaseStorage["key1.key2"] = "value2"
    assert BaseStorage.contains_item("key1.key2", "value2")

    del BaseStorage["key1"]

    assert not BaseStorage.contains_item("key1.key2.key3.key4.key5.key6", "values")

    BaseStorage["key1.key2.key3.key4.key5.key6"] = "values"
    assert BaseStorage.contains_item("key1.key2.key3.key4.key5.key6", "values")

def test_basestorage_contains_item_keytype():
    """Test the Storage.contains_item() methods' protection against incorrect key types"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(WrongTypeError):
        BaseStorage.contains_item(None, "value")
    with pytest.raises(WrongTypeError):
        BaseStorage.contains_item(10, "value")
    with pytest.raises(WrongTypeError):
        BaseStorage.contains_item(list(("1",)), "value")

    with pytest.raises(InvalidKeyError):
        BaseStorage.contains_item(".some.key", "value")
    with pytest.raises(InvalidKeyError):
        BaseStorage.contains_item("some.key.", "value")
    with pytest.raises(InvalidKeyError):
        BaseStorage.contains_item("some..key", "value")

def test_basestorage_contains_item_valuetype():
    """Test the Storage.contains_item() methods' support for different value types"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert not BaseStorage.contains_item("key1", ["1", 2, None])

    BaseStorage["key1"] = ["1", 2, None]
    assert BaseStorage.contains_item("key1", ["1", 2, None])

def test_basestorage_get():
    """Test the Storage.get() method"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert not BaseStorage.contains("key1")
    assert BaseStorage.get("key1") is None

    BaseStorage["key1"] = "value"
    assert BaseStorage.contains("key1")
    assert BaseStorage.get("key1") == "value"

def test_basestorage_get_multi():
    """Test the Storage.get() method with subdicts specified in the key"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert not BaseStorage.contains("key1.key2")
    assert BaseStorage.get("key1.key2") is None

    BaseStorage["key1.key2"] = "value2"
    assert BaseStorage.contains("key1.key2")
    assert BaseStorage.get("key1.key2") == "value2"

    del BaseStorage["key1"]

    assert not BaseStorage.contains("key1.key2.key3.key4.key5.key6")
    assert BaseStorage.get("key1.key2.key3.key4.key5.key6") is None

    BaseStorage["key1.key2.key3.key4.key5.key6"] = "values"
    assert BaseStorage.get("key1.key2.key3.key4.key5.key6") == "values"

def test_basestorage_get_keytype():
    """Test the Storage.get() methods' protection against incorrect key types"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(WrongTypeError):
        BaseStorage.get(None)
    with pytest.raises(WrongTypeError):
        BaseStorage.get(10)
    with pytest.raises(WrongTypeError):
        BaseStorage.get(list(("1",)))

    with pytest.raises(InvalidKeyError):
        BaseStorage.get(".some.key")
    with pytest.raises(InvalidKeyError):
        BaseStorage.get("some.key.")
    with pytest.raises(InvalidKeyError):
        BaseStorage.get("some..key")

def test_basestorage_get_default():
    """Test the Storage.get() methods' optional default argument"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert not BaseStorage.contains("key1")
    assert BaseStorage.get("key1") is None

    assert BaseStorage.get("key1", None) is None
    assert BaseStorage.get("key1", default=None) is None

    assert BaseStorage.get("key1", 45) == 45
    assert BaseStorage.get("key1", default=45) == 45

    assert BaseStorage.get("key1", "test") == "test"
    assert BaseStorage.get("key1", default="test") == "test"

    BaseStorage["key1"] = "value"
    assert BaseStorage.contains("key1")
    assert BaseStorage.get("key1") == "value"

    assert BaseStorage.get("key1", None) == "value"
    assert BaseStorage.get("key1", default=None) == "value"

    assert BaseStorage.get("key1", 45) == "value"
    assert BaseStorage.get("key1", default=45) == "value"

    assert BaseStorage.get("key1", "test") == "value"
    assert BaseStorage.get("key1", default="test") == "value"

def test_basestorage_name():
    """Test the Storage.name member"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    assert hasattr(BaseStorage, "name")
    assert isinstance(BaseStorage.name, str)
    assert BaseStorage.name == "BaseStorage"

def test_basestorage_name_readonly():
    """Ensure that Storage.name is readonly"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(ReadonlyError):
        BaseStorage.name = "AnotherName"

def test_basestorage_name_custom():
    """Ensure that custom storages need to overwrite _STORAGE_NAME"""

    BaseStorage = _BaseStorage.__new__(_BaseStorage)
    BaseStorage._store = {}

    with pytest.raises(UninitializedFieldError):
        class _TestStorage(_BaseStorage):
            def __init__(self):
                pass

    with pytest.raises(WrongTypeError):
        class _TestStorage2(_BaseStorage):
            def __init__(self):
                pass
            _STORAGE_NAME = 42

    with pytest.raises(UninitializedFieldError):
        class _TestStorage3(_BaseStorage):
            def __init__(self):
                pass
            _STORAGE_NAME = "BaseStorage"

    class _TestStorage4(_BaseStorage):
        def __init__(self):
            pass
        _STORAGE_NAME = "TestStorage4"

def test_internalstorage_inheritance():
    """Ensure the InternalStorage inherits from _BaseStorage"""

    InternalStorage = _InternalStorage.__new__(_InternalStorage)
    InternalStorage._store = {}

    assert isinstance(InternalStorage, _BaseStorage)

def test_internalstorage_instantiation():
    """Ensure that InternalStorage behaves like a singleton"""

    with pytest.raises(SingletonInstantiationError):
        _InternalStorage()._init()

    with pytest.raises(SingletonInstantiationError):
        _InternalStorage()._init()

def test_internalstorage_setitem():
    """Ensure that InternalStorage only accepts keys prefixed with '__'"""

    InternalStorage = _InternalStorage.__new__(_InternalStorage)
    InternalStorage._store = {}

    with pytest.raises(InternalFunctionUsedError):
        InternalStorage["key1"] = "value"
    with pytest.raises(InternalFunctionUsedError):
        InternalStorage["key1_"] = "value"
    with pytest.raises(InternalFunctionUsedError):
        InternalStorage["key__1"] = "value"

    InternalStorage["_key0"] = "value1"
    assert InternalStorage._store["_key0"] == "value1"

    InternalStorage["_key1._key2"] = "value2"
    assert InternalStorage._store["_key1"]["_key2"] == "value2"

    InternalStorage["_key2.key2"] = "value3"
    assert InternalStorage._store["_key2"]["key2"] == "value3"



"""Module containing custom exceptions for general usage"""

# pylint: disable=arguments-differ

from typing import Any

from ._custom_exception import CustomException

class CalledMultipleTimesError(CustomException):
    """Exception raised when a single-use function is called twice"""

    default_messages = {
        0: "The function can only be called once"
    }

class DeprecatedError(CustomException):
    """Exception raised when deprecated functionality is used"""

    default_messages = {
        0: "This functionality is deprecated",
        1: "This functionality is deprecated, please use {0} instead"
    }

class DirNotFoundError(CustomException):
    """Exception raised when an expected directory doesn't exist"""

    default_messages = {
        0: "The expected directory doesn't exist",
        1: "The expected directory '{0}' doesn't exist"
    }

class InternalCalculationError(CustomException):
    """Exception raised when an internal calculation resulted in an unacceptable value"""

    default_messages = {
        0: "Internal calculation resulted in an unacceptable value",
        1: "Internal calculation resulted in an unacceptable value {0}"
    }

class InternalFunctionUsedError(CustomException):
    """Exception raised when an internal function was used by an external project"""

    default_messages = {
        0: "This function is only for library-internal use"
    }

class InvalidKeyError(CustomException):
    """Exception raised when the key has an invalid format"""

    default_messages = {
        0: "The key has an invalid format",
        1: "The key '{0}' has an invalid format"
    }

class KeyNotFoundError(CustomException):
    """Exception raised when the key is not found in the storage"""

    default_messages = {
        0: "The requested key could not be found",
        1: "The requested key '{0}' could not be found"
    }

class LockAcquisitionTimeoutError(CustomException):
    """Exception raised when the acquisitgion of a lock timed out"""

    default_messages = {
        0: "The requested lock could not be acquired in time"
    }

class MissingDefaultMessageError(CustomException):
    """Exception raised when an error class is missing the default message"""

    default_messages = {
        0: "The error class is missing a default message. Set it as a class variable in default_messages[0].",
        1: "The error class '{0}' is missing a default message. Set it as a class variable in default_messages[0]."
    }

class MissingInheritanceError(CustomException):
    """Exception raised when a class is expected to inherit from another class"""

    default_messages = {
        0: "The class is missing an inheritance from another class",
        1: "The class is missing an inheritance from {0}",
        2: "The class {1} is missing an inheritance from {0}"
    }

    @classmethod
    def with_values(cls, class_name: Any | type, base_class_name: Any | type):
        if not isinstance(class_name, type):
            class_name = type(class_name)
        if not isinstance(base_class_name, type):
            base_class_name = type(base_class_name)

        return super().with_values(base_class_name, class_name)

class NameNotFoundError(CustomException):
    """Exception raised when the name is not found"""

    default_messages = {
        0: "The requested name could not be found",
        1: "The requested name '{0}' could not be found"
    }

class NoneTypeError(CustomException):
    """Exception raised when a value is unexpectedly None"""

    default_messages = {
        0: "Didn't expect None as a value here"
    }

class NotInitializedError(CustomException):
    """Exception raised when an instance is used before it is correctly initialized"""

    default_messages = {
        0: "The instance is not yet initialized correctly"
    }

class RegisteredMultipleTimesError(CustomException):
    """Exception raised when something is registered twice"""

    default_messages = {
        0: "This object is already registered",
        1: "{0} is already registered"
    }

class ReadonlyError(CustomException):
    """Exception raised when a read-only object is changed"""

    default_messages = {
        0: "This object is read-only and cannot be changed",
        1: "{0} is read-only and cannot be changed"
    }

class SingletonInstantiationError(CustomException):
    """Exception raised when a singleton class is instantiated twice"""

    default_messages = {
        0: "The singleton class can only be instantiated once",
        1: "The singleton class {0} can only be instantiated once"
    }

    @classmethod
    def with_values(cls, class_name: Any | type):
        if not isinstance(class_name, type):
            class_name = type(class_name)

        return super().with_values(class_name)

class UninitializedFieldError(CustomException):
    """Exception raised when a subclass doesn't initialize a mandatory field"""

    default_messages = {
        0: "The subclass doesn't initialize a mandatory field",
        1: "The subclass doesn't initialize the mandatory field {0}",
        2: "{0} doesn't initialize the mandatory field {1}"
    }

class WrongTypeError(CustomException):
    """Exception raised when a value wasn't of an expected type"""

    default_messages = {
        0: "Received an unexpected type",
        2: "Expected {1}, not {0}",
        3: "Expected {1} or {2}, not {0}",
        4: "Expected {1}, {2} or {3}, not {0}",
        5: "Expected {1}, {2}, {3} or {4}, not {0}",
        6: "Expected {1}, {2}, {3}, {4} or {5}, not {0}",
    }

    @classmethod
    def with_values(cls, received: Any | type, *expected: Any | type):
        if not isinstance(received, type):
            received = type(received)
        expected = list(expected)
        for c, item in enumerate(expected):
            if not isinstance(item, type):
                expected[c] = type(item)

        return super().with_values(received, *expected)



"""A module containing the levenshtein_distance function"""

import numpy

from ..error import WrongTypeError

def levenshtein_distance(token1: str, token2: str) -> int:
    """
    Calculate the levenshtein distance between token1 and token2

    This represents the edit distance between two strings
    """

    if not isinstance(token1, str):
        raise WrongTypeError.with_values(token1, str)
    if not isinstance(token2, str):
        raise WrongTypeError.with_values(token2, str)

    distances = numpy.zeros((len(token1) + 1, len(token2) + 1), dtype=int)

    for t1 in range(len(token1) + 1):
        distances[t1][0] = t1

    for t2 in range(len(token2) + 1):
        distances[0][t2] = t2

    a = 0
    b = 0
    c = 0

    for t1 in range(1, len(token1) + 1):
        for t2 in range(1, len(token2) + 1):
            if token1[t1-1] == token2[t2-1]:
                distances[t1][t2] = distances[t1 - 1][t2 - 1]
            else:
                a = distances[t1][t2 - 1]
                b = distances[t1 - 1][t2]
                c = distances[t1 - 1][t2 - 1]

                if (a <= b and a <= c):
                    distances[t1][t2] = a + 1
                elif (b <= a and b <= c):
                    distances[t1][t2] = b + 1
                else:
                    distances[t1][t2] = c + 1

    final_dist = distances[len(token1)][len(token2)]
    # cast np.int64 to int
    return final_dist.item()



"""Module containing the _CacheStorage class"""

from .._storage._base_storage import _BaseStorage
from .. import error

class _CacheStorage(_BaseStorage):
    """Storage used for caching values"""

    def __init__(self) -> None:
        if _CacheStorage._instance is not None:
            raise error.SingletonInstantiationError.with_values(_CacheStorage)

        _CacheStorage._instance = self
        _CacheStorage._store = self._store = {}

    _STORAGE_NAME = "CacheStorage"

    def _ensure_initialized(self):
        # the storage is always initialized
        return True



"""A module containing path-modification functions."""

import pathlib

def absolute(*paths: str | pathlib.Path) -> str:
    """
    Return an absolute path, regardless of what is input.

    Additionally, the path is resolved, removing any symlinks on the way.
    """

    if len(paths) == 0:
        raise ValueError()
    for item in paths:
        if not isinstance(item, (str, pathlib.Path)):
            raise TypeError()

    path = pathlib.Path(*paths)

    if path.is_absolute():
        return str(path.resolve())

    return str(path.absolute().resolve())



"""A module containing the WorkerThread class"""

from threading import Thread
from typing import Any

# original code from https://stackoverflow.com/a/6894023
class WorkerThread(Thread):
    """Wrapper around `threading.Thread` that stores and returns resulting values and exceptions on join."""

    def __init__(self,
                 group=None,
                 target=None,
                 name=None,
                 args=(),
                 kwargs=None,
                 daemon=None):
        super().__init__(group, target, name, args, kwargs, daemon=daemon)

        self._return = None

    def run(self) -> None:
        """Invoke the callable object."""

        if self._target is not None:
            try:
                self._return = self._target(*self._args, **self._kwargs)
            # pylint: disable-next=broad-exception-caught
            except BaseException as e:
                self._return = e

    def join(self, timeout: float | None = None, reraise: bool = False) -> Any | BaseException:
        """Wait until the thread terminates and return any stored values."""

        super().join(timeout)

        if reraise and isinstance(self._return, BaseException):
            raise self._return

        return self._return

    def failed(self) -> bool:
        """Return whether target execution raised an exception."""

        return isinstance(self._return, BaseException)



"""
    Pytest configuration
"""

# pylint: disable=wrong-import-position, wrong-import-order, missing-function-docstring

# Adding source path to sys path
import pathlib
import sys
import os

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), '../'))
sys.path.append(f"{pathlib.Path(__file__).parent.parent}")
sys.path.append(f"{pathlib.Path(__file__).parent}")
# pylint: enable=wrong-import-position, wrong-import-order

import pytest

os.environ["DEBUG"] = "True"

def pytest_addoption(parser):
    parser.addoption(
        "--skip-linting", action="store_true", default=False, help="skip the pylint test"
    )

def pytest_collection_modifyitems(config, items):
    if config.getoption("--skip-linting"):
        skip_pylint = pytest.mark.skip(reason="skipping code linting due to --skip-linting arg")
        for item in items:
            if item.name == "test_pylint":
                item.add_marker(skip_pylint)

# pylint: disable-next=unused-wildcard-import, wildcard-import
from .fixtures import *



"""Module containing the CustomException class"""

from typing import Any

class CustomException(Exception):
    """
    The base class for all custom exceptions
    
    If no arguments are provided at instantiation, the default error message is used.
    Otherwise, the provided argument is used as the error message.
    """

    def __init__(self, *args, **kwargs):
        if self.__class__ == CustomException:
            raise NotImplementedError()

        if args:
            if not isinstance(args[0], str):
                # needs to be imported here to prevent circular import
                # pylint: disable-next=cyclic-import, import-outside-toplevel
                from abllib.error._general import WrongTypeError
                raise WrongTypeError("Expected error message to be of type str")
            super().__init__(*args, **kwargs)
        else:
            # exception was raised without args
            super().__init__(self.default_messages[0], **kwargs)

    default_messages: dict[int, str]

    @classmethod
    def with_values(cls, *args: Any):
        """Instantiate the given exception with the given args"""

        if cls == CustomException:
            raise NotImplementedError()

        index = len(args)
        if index not in cls.default_messages:
            raise AttributeError(f"{cls} does not support error messages created with {index} arguments")

        message = cls.default_messages[index]
        message = message.format(*args)
        return cls(message)

    def __init_subclass__(cls):
        if 0 not in cls.default_messages:
            # needs to be imported here to prevent circular import
            # pylint: disable-next=cyclic-import, import-outside-toplevel
            from abllib.error._general import MissingDefaultMessageError
            raise MissingDefaultMessageError.with_values(cls)

        for key in cls.default_messages.keys():
            if not isinstance(key, int):
                raise TypeError()
        for val in cls.default_messages.values():
            if not isinstance(val, str):
                raise TypeError()

        if hasattr(cls, "default_message"):
            # TODO: handle legacy classes, as default_message is deprecated
            pass



"""A module containing the logger creation"""

import atexit
import logging
import sys
from enum import Enum
from typing import Literal

from . import error, fs
from ._storage import InternalStorage

DEFAULT_LOG_LEVEL = logging.INFO

class LogLevel(Enum):
    """An enum holding log levels"""

    CRITICAL = logging.CRITICAL
    FATAL = logging.FATAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    WARN = logging.WARN
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    ALL = 1
    NOTSET = logging.NOTSET

    def __eq__(self, other):
        return self is other or self.value == other

    def __ne__(self, other):
        return self is not other and self.value != other

    # for more details look here:
    # https://stackoverflow.com/a/72664895/15436169
    def __hash__(self):
        return hash(self.value)

def initialize(log_level: Literal[LogLevel.CRITICAL]
                          | Literal[LogLevel.ERROR]
                          | Literal[LogLevel.WARNING]
                          | Literal[LogLevel.INFO]
                          | Literal[LogLevel.DEBUG]
                          | Literal[LogLevel.ALL]
                          | None = None):
    """
    Initialize the custom logging module.

    This disables all log output. Use the add_<*>_handler functions to complete the setup.

    This function removes any previous logging setup, also overwriting the root logger formatter.
    """

    logging.disable()

    # remove existing handlers
    if "_log.handlers" in InternalStorage:
        for handler in InternalStorage["_log.handlers"]:
            get_logger().removeHandler(handler)

            # remove atexit function
            if isinstance(handler, logging.FileHandler):
                atexit.unregister(handler.close)
                handler.close()

    if log_level is None:
        InternalStorage["_log.level"] = DEFAULT_LOG_LEVEL
        get_logger().setLevel(DEFAULT_LOG_LEVEL)
        return

    if not isinstance(log_level, (int, LogLevel)):
        raise TypeError(f"Expected log_level to be of type {int | LogLevel}, but got {type(log_level)}")

    if isinstance(log_level, LogLevel):
        log_level = log_level.value

    if log_level == LogLevel.NOTSET:
        raise ValueError("LogLevel.NOTSET is not allowed.")
    if log_level == LogLevel.WARN:
        raise error.DeprecatedError("LogLevel.WARN is deprecated, use LogLevel.WARNING instead.")
    if log_level == LogLevel.FATAL:
        raise ValueError("Loglevel.FATAL should not be used, use LogLevel.CRITICAL instead.")

    InternalStorage["_log.level"] = log_level
    get_logger().setLevel(log_level)

def add_console_handler() -> None:
    """
    Add a console handler to the root logger.

    This configures all loggers to also print to sys.stdout.
    """

    if "_log.level" not in InternalStorage:
        raise error.NotInitializedError("log.initialize() needs to be called first")

    logging.disable(0)

    stream_handler = logging.StreamHandler(sys.stdout)

    stream_handler.setLevel(InternalStorage["_log.level"])

    stream_handler.setFormatter(_get_formatter())

    get_logger().addHandler(stream_handler)

    # add logger to storage
    if "_log.handlers" not in InternalStorage:
        InternalStorage["_log.handlers"] = []
    InternalStorage["_log.handlers"].append(stream_handler)

def add_file_handler(filename: str = "latest.log") -> None:
    """
    Add a file handler to the root logger.

    This configures all loggers to also print to a given file, or 'latest.log' if not provided.
    """

    if "_log.level" not in InternalStorage:
        raise error.NotInitializedError("log.initialize() needs to be called first")

    logging.disable(0)

    file_handler = logging.FileHandler(filename=fs.absolute(filename), encoding="utf-8", mode="w")

    file_handler.setLevel(InternalStorage["_log.level"])

    file_handler.setFormatter(_get_formatter())

    get_logger().addHandler(file_handler)

    atexit.register(file_handler.close)

    # add logger to storage
    if "_log.handlers" not in InternalStorage:
        InternalStorage["_log.handlers"] = []
    InternalStorage["_log.handlers"].append(file_handler)

def get_logger(name: str = None) -> logging.Logger:
    """
    Return a logger with the given name, or the root logger if name is None.

    If a logger doesn't yet exist, it is created and then returned.
    """

    if name is None:
        return logging.getLogger()

    if not isinstance(name, str):
        raise TypeError(f"Expected logger name to be of type {str}, but got {type(name)}")

    return logging.getLogger(name)

def get_loglevel() -> Literal[LogLevel.CRITICAL] \
                      | Literal[LogLevel.ERROR] \
                      | Literal[LogLevel.INFO] \
                      | Literal[LogLevel.WARNING] \
                      | Literal[LogLevel.DEBUG] \
                      | Literal[LogLevel.ALL] \
                      | None:
    """Return the current LogLevel"""

    return InternalStorage["_log.level"] if "_log.level" in InternalStorage else None

def _get_formatter():
    dt_fmt = r"%Y-%m-%d %H:%M:%S"
    formatter = logging.Formatter("[{asctime}] [{levelname:<8}] {name}: {message}", dt_fmt, style="{")
    return formatter



"""A module containing the WorkerProcess class"""

# pylint: disable=dangerous-default-value

from multiprocessing import Process, Queue
from time import sleep
from typing import Any

import dill

from abllib.log import get_logger

class WorkerProcess(Process):
    """Wrapper around `multiprocessing.Process` that stores and returns resulting values and exceptions."""

    def __init__(self,
                 group=None,
                 target=None,
                 name=None,
                 args=(),
                 kwargs={},
                 daemon=None):
        super().__init__(group, target, name, args, kwargs, daemon=daemon)

        # use dill instead of pickle
        # https://stackoverflow.com/a/72776044/15436169
        self._target = dill.dumps(self._target)
        self._return_queue = Queue(maxsize=1)
        self._failed_queue = Queue(maxsize=1)

    def run(self) -> None:
        """Invoke the callable object."""

        if self._target is not None:
            try:
                target = dill.loads(self._target)
                return_value = target(*self._args, **self._kwargs)
                self._return_queue.put(return_value)
            # pylint: disable-next=broad-exception-caught
            except BaseException as e:
                self._return_queue.put(e)
                self._failed_queue.put(None)

        else:
            self._return_queue.put(None)

    def join(self, timeout: float | None = None, reraise: bool = False) -> Any | BaseException:
        """Wait until the child process terminates."""

        super().join(timeout)

        return_value = self._return_queue.get(block=False)

        # put return_value back to allow multiple .join() calls
        self._return_queue.put(return_value)

        # wait until the value is actually in the queue again
        # https://stackoverflow.com/a/63355643/15436169
        c = 0
        while self._return_queue.empty():
            c += 1

            sleep(0.001 * c)

            if c == 10:
                # after testing, sleeping for 1 ms takes ~1.9 ms real-time
                # so the overhead of time.sleep is ~0.9 ms
                get_logger("abllib").debug(f"queue.put not complete after ~{sum(range(c + 1)) + (0.9 * c)} ms")
                break

        if reraise and isinstance(return_value, BaseException):
            raise return_value

        return return_value

    def failed(self) -> bool:
        """Return whether target execution raised an exception."""

        return not self._failed_queue.empty()



"""A module containing a custom Lock and Semaphore class"""

import threading

class Lock():
    """
    Extends threading.Lock by allowing timeout to be None.

    threading.Lock cannot be subclassed as it is a factory function.
    https://stackoverflow.com/a/6781398
    """

    def __init__(self):
        self._lock = threading.Lock()

    def acquire(self, blocking: bool = True, timeout: float | None = None):
        """
        Try to acquire the Lock.
        
        If blocking is disabled, it doesn't wait for the timeout.

        If timeout is set, wait for n seconds before returning.
        """

        return self._lock.acquire(blocking, -1 if timeout is None else timeout)

    def locked(self) -> bool:
        """Returns whether the Lock is held"""

        return self._lock.locked()

    def release(self):
        """Release the lock if it is currently held"""

        self._lock.release()

    def __enter__(self):
        self.acquire()

    # keep signature the same as threading.Lock
    # pylint: disable-next=redefined-builtin
    def __exit__(self, type, value, traceback):
        self.release()

# we can't use the default threading.Semaphore
# because we need a semaphore with value == 0 if it isn't held
# This is the opposite behaviour of threading.Semaphore
class Semaphore(threading.BoundedSemaphore):
    """
    Extends threading.BoundedSemaphore by adding a locked() function.

    This makes it equivalent to threading.Lock method signature-wise.
    """

    def locked(self) -> bool:
        """Returns whether the Semaphore is held at least once"""

        return self._value != self._initial_value
